from pathlib import Path
from typing import Callable, Iterable, List, Optional, Union, cast

import spacy
from spacy.language import Language
from spacy.tokens import Doc
from spacy.training import Example
from spacy.util import SimpleFrozenList
import srsly

from pymusas.file_utils import ensure_path
from pymusas.rankers.lexicon_entry import LexiconEntryRanker
from pymusas.rankers.ranking_meta_data import RankingMetaData
from pymusas.spacy_api.utils import set_custom_token_extension
from pymusas.taggers.rules.rule import Rule


class RuleBasedTagger(spacy.pipeline.pipe.Pipe):
    '''
    [spaCy pipeline component](https://spacy.io/usage/processing-pipelines)
    of the :class:`pymusas.taggers.rule_based.RuleBasedTagger`.

    This component applies one or more :class:`pymusas.taggers.rules.rule.Rule`s
    to create a list of possible candidate tags for each token in the sequence.
    Each candidate, represented as a
    :class:`pymusas.rankers.ranking_meta_data.RankingMetaData` object, for each
    token is then Ranked using a
    :class:`pymusas.rankers.lexicon_entry.LexiconEntryRanker` ranker. The best
    candidate and it's associated tag(s) for each token are then assigned to the
    `Token._.pymusas_tags` attribute in addition a `List` of token indexes
    indicating if the token is part of a Multi Word Expression (MWE) is assigned
    to the `Token._.pymusas_mwe_indexes`.

    If we cannot tag a token then the following process will happen:
    1. If the token's POS tag is in `default_punctuation_tags` then it will assign the
    tag `PUNCT`.
    2. If the token's POS tag is in `default_number_tags` then it will assign the tag
    `N1`.
    3. Assign the default tag `Z99`.

    **NOTE** this tagger has been designed to be flexible with the amount of
    resources avaliable, for example if you do not have a POS tagger or
    lemmatiser in your spaCy pipeline this ok, just use the default
    `pos_attribute` and `lemma_attribute`.

    # Assigned Attributes

    <table>
        <tr>
            <th> Location </th>
            <th> Type </th>
            <th> Value </th>
        </tr>
        <tr>
            <td> Token._.pymusas_tags </td>
            <td> `List[str]` </td>
            <td> Prediced tags, the first tag in the List of tags is the
            most likely tag. </td>
        </tr>
        <tr>
            <td> Token._.pymusas_mwe_indexes </td>
            <td> `List[Tuple[int, int]]` </td>
            <td> Each `Tuple` indicates the start and end token index of the
            associated Multi Word Expression (MWE). If the `List` contains
            more than one `Tuple` then the MWE is discontinuous. For single word
            expressions the `List` will only contain 1 `Tuple` which will be
            (token_start_index, token_start_index + 1).
            </td>
        </tr>
    </table>
    
    # Config and implementation

    The default config is defined by the pipeline component factory and describes
    how the component should be configured. You can override its settings via the `config`
    argument on [nlp.add_pipe](https://spacy.io/api/language#add_pipe) or in your
    [config.cfg for training](https://spacy.io/usage/training#config).

    | Setting                  | Description                  |
    |--------------------------|------------------------------|
    | pymusas_tags_token_attr  | See parameters section below |
    | pymusas_mwe_indexes_attr | See parameters section below |
    | pos_attribute            | See parameters section below |
    | lemma_attribute          | See parameters section below |

    # Parameters

    name : `str`, optional (default = `pymusas_rule_based_tagger`)
        The component name. Defaults to the same name as the class variable
        `COMPONENT_NAME`.
    pymusas_tags_token_attr : `str`, optional (default = `pymusas_tags`)
        The name of the attribute to assign the predicted tags too under
        the `Token._` class.
    pymusas_mwe_indexes_attr : `str`, optional (default = `pymusas_mwe_indexes`)
        The name of the attribute to assign the start and end token index of the
        associated MWE too under the `Token._` class.
    pos_attribute : `str`, optional (default = `pos_`)
        The name of the attribute that the Part Of Speech (POS) tag is assigned too
        within the `Token` class. The POS tag value that comes from this attribute
        has to be of type `str`. With the current default we take the POS tag
        from `Token.pos_`. The POS tag can be an empty string if you do not require
        POS information or if you do not have a POS tagger. **NOTE** that if you
        do not have a POS tagger the default value for `Token.pos_` is an empty
        string.
    lemma_attribute : `str`, optional (default = `lemma_`)
        The name of the attribute that the lemma is assigned too within the `Token`
        class. The lemma value that comes from this attribute has to be of
        type `str`. With the current default we take the lemma from `Token.lemma_`.
        The lemma can be an empty string if you do not require
        lemma information or if you do not have a lemmatiser. **NOTE** that if you
        do not have a lemmatiser the default value for `Token.lemma_` is an empty
        string.

    # Instance Attributes

    name : `str`
        The component name.
    rules : `List[pymusas.taggers.rules.rule.Rule]`, optional (default = `None`)
        The `rules` is set through the :func:`initialize` method. Before it is
        set by the :func:`initialize` method the value of this attribute is `None`.
    ranker : `pymusas.rankers.lexicon_entry.LexiconEntryRanker`, optional (default = `None`)
        The `ranker` is set through the :func:`initialize` method. Before it is
        set by the :func:`initialize` method the value of this attribute is `None`.
    default_punctuation_tags : `Set[str]`
        The `default_punctuation_tags` is set through the :func:`initialize` method.
    default_number_tags : `Set[str]`
        The `default_number_tags` is set through the :func:`initialize` method.
    pymusas_tags_token_attr : `str`, optional (default = `pymusas_tags`)
        The given `pymusas_tags_token_attr`
    pymusas_mwe_indexes_attr : `str`, optional (default = `pymusas_mwe_indexes`)
        The given `pymusas_mwe_indexes_attr`
    pos_attribute : `str`, optional (default = `pos_`)
        The given `pos_attribute`
    lemma_attribute : `str`, optional (default = `lemma_`)
        The given `lemma_attribute`

    # Class Attributes

    COMPONENT_NAME : `str`
        Name of component factory that this component is registered under. This
        is used as the first argument to
        [`Language.add_pipe`](https://spacy.io/api/language#add_pipe)
        if you want to add this component to your spaCy pipeline.

    # Examples

    ``` python
    >>> import spacy
    >>> from pymusas.rankers.lexicon_entry import ContextualRuleBasedRanker
    >>> from pymusas.pos_mapper import BASIC_CORCENCC_TO_USAS_CORE
    >>> from pymusas.lexicon_collection import LexiconCollection
    >>> from pymusas.taggers.rules.single_word import SingleWordRule
    >>> from pymusas.spacy_api.taggers.rule_based import RuleBasedTagger
    >>> # Construction via spaCy pipeline
    >>> nlp = spacy.blank('en')
    >>> # Using default config
    >>> single_lexicon_url = 'https://raw.githubusercontent.com/UCREL/Multilingual-USAS/master/Welsh/semantic_lexicon_cy.tsv'
    >>> single_lexicon = LexiconCollection.from_tsv(single_lexicon_url)
    >>> single_lemma_lexicon = LexiconCollection.from_tsv(single_lexicon_url,
    ...                                                   include_pos=False)
    >>> single_rule = SingleWordRule(single_lexicon, single_lemma_lexicon,
    ...                              pos_mapper=BASIC_CORCENCC_TO_USAS_CORE)
    >>> rules = [single_rule]
    >>> ranker = ContextualRuleBasedRanker(*ContextualRuleBasedRanker.get_construction_arguments(rules))
    >>> tagger = nlp.add_pipe('pymusas_rule_based_tagger')
    >>> tagger.rules = rules
    >>> tagger.ranker = ranker
    >>> token = nlp('aberth')
    >>> assert token[0]._.pymusas_tags == ['S9', 'A9-']
    >>> assert token[0]._.pymusas_mwe_indexes == [(0, 1)]
    >>> # Custom config
    >>> custom_config = {'pymusas_tags_token_attr': 'semantic_tags',
    ...                  'pymusas_mwe_indexes_attr': 'mwe_indexes'}
    >>> nlp = spacy.blank('en')
    >>> tagger = nlp.add_pipe('pymusas_rule_based_tagger', config=custom_config)
    >>> tagger.rules = rules
    >>> tagger.ranker = ranker
    >>> token = nlp('aberth')
    >>> assert token[0]._.semantic_tags == ['S9', 'A9-']
    >>> assert token[0]._.mwe_indexes == [(0, 1)]

    ```
    '''
    COMPONENT_NAME = 'pymusas_rule_based_tagger'

    def __init__(self,
                 name: str = 'pymusas_rule_based_tagger',
                 pymusas_tags_token_attr: str = 'pymusas_tags',
                 pymusas_mwe_indexes_attr: str = 'pymusas_mwe_indexes',
                 pos_attribute: str = 'pos_',
                 lemma_attribute: str = 'lemma_'
                 ) -> None:
        self.name = name
        
        self._pymusas_tags_token_attr = pymusas_tags_token_attr
        set_custom_token_extension(pymusas_tags_token_attr)

        self._pymusas_mwe_indexes_attr = pymusas_mwe_indexes_attr
        set_custom_token_extension(pymusas_mwe_indexes_attr)
        
        self._pos_attribute = pos_attribute
        self._lemma_attribute = lemma_attribute
        
        self.rules: Optional[List[Rule]] = None
        self.ranker: Optional[LexiconEntryRanker] = None
        
        self.default_punctuation_tags = set(['punc'])
        self.default_number_tags = set(['num'])
        
        self._validated = False

    def _validate(self) -> None:
        '''
        Checks that `rules` and `ranker` are not `None`
        '''
        error_msg = ('The `{}` attribute cannot be `None`, this '
                     'attribute can be set through the `initialize` method.')
        if self.rules is None:
            raise ValueError(error_msg.format('rules'))
        
        if self.ranker is None:
            raise ValueError(error_msg.format('ranker'))

        self._validated = True
    
    def initialize(self,
                   get_examples: Optional[Callable[[], Iterable[Example]]] = None,
                   *,
                   nlp: Optional[Language] = None,
                   rules: Optional[List[Rule]] = None,
                   ranker: Optional[LexiconEntryRanker] = None,
                   default_punctuation_tags: Optional[List[str]] = None,
                   default_number_tags: Optional[List[str]] = None,
                   ) -> None:
        '''
        Initialize the tagger and load any of the resources given. The method is
        typically called by
        [`Language.initialize`](https://spacy.io/api/language#initialize)
        and lets you customize arguments it receives via the
        [`initialize.components`](https://spacy.io/api/data-formats#config-initialize)
        block in the config. The loading only happens during initialization,
        typically before training. At runtime, all data is load from disk.

        # Parameters
        
        rules : `List[pymusas.taggers.rules.rule.Rule]`
            A list of rules to apply to the sequence of tokens in the
            :func:`__call__`. The output from each rule is concatendated and given
            to the `ranker`.
        ranker : `pymusas.rankers.lexicon_entry.LexiconEntryRanker`
            A ranker to rank the output from all of the `rules`.
        default_punctuation_tags : `List[str]`, optional (default = `None`)
            The POS tags that represent punctuation. If `None` then we will use
            `['punc']`. The list will be converted into a `Set` before assigning
            to the `default_punctuation_tags` attribute.
        default_number_tags : `List[str]`, optional (default = `None`)
            The POS tags that represent numbers. If `None` then we will use
            `['num']`. The list will be converted into a `Set` before assigning
            to the `default_number_tags` attribute.
        '''
        if rules is not None:
            self.rules = rules
        
        if ranker is not None:
            self.ranker = ranker
        
        if default_punctuation_tags is not None:
            self.default_punctuation_tags = set(default_punctuation_tags)
        
        if default_number_tags is not None:
            self.default_number_tags = set(default_number_tags)
        
        self._validate()
    
    def __call__(self, doc: Doc) -> Doc:
        '''
        Applies the tagger to the spaCy document, modifies it in place, and
        returns it. This usually happens under the hood when the `nlp` object is
        called on a text and all pipeline components are applied to the `Doc` in
        order.
        
        # Parameters

        doc : `Doc`
            A [spaCy `Doc`](https://spacy.io/api/doc)

        # Returns

        `Doc`
        '''
        if not self._validated:
            self._validate()
        
        ranker = cast(LexiconEntryRanker, self.ranker)
        rules = cast(List[Rule], self.rules)
        
        # Try, catch error handling reference:
        # https://github.com/explosion/spaCy/blob/6af6c2e86cc7b08573b261563786bd1ab87d45e9/spacy/pipeline/lemmatizer.py#L131
        error_handler = self.get_error_handler()
        try:
            tokens: List[str] = []
            lemmas: List[str] = []
            pos_tags: List[str] = []
            for token in doc:
                tokens.append(token.text)
                lemmas.append(getattr(token, self.lemma_attribute))
                pos_tags.append(getattr(token, self.pos_attribute))
            token_ranking_meta_data: List[List[RankingMetaData]] \
                = [[] for _ in range(len(tokens))]
            for rule in rules:
                rule_ranking_meta_data = rule(tokens, lemmas, pos_tags)
                for token_index, ranking_meta_data in enumerate(rule_ranking_meta_data):
                    token_ranking_meta_data[token_index].extend(ranking_meta_data)
            
            token_ranks, token_best_rank = ranker(token_ranking_meta_data)
            
            for token_index, best_rank in enumerate(token_best_rank):
                token = doc[token_index]
                if best_rank is None:
                    pos_tag = pos_tags[token_index]
                    if pos_tag in self.default_punctuation_tags:
                        setattr(token._, self.pymusas_tags_token_attr, ['PUNCT'])
                        setattr(token._, self.pymusas_mwe_indexes_attr,
                                [(token_index, token_index + 1)])
                    elif pos_tag in self.default_number_tags:
                        setattr(token._, self.pymusas_tags_token_attr, ['N1'])
                        setattr(token._, self.pymusas_mwe_indexes_attr,
                                [(token_index, token_index + 1)])
                    else:
                        setattr(token._, self.pymusas_tags_token_attr, ['Z99'])
                        setattr(token._, self.pymusas_mwe_indexes_attr,
                                [(token_index, token_index + 1)])
                    continue
                tags = list(best_rank.semantic_tags)
                indexes = [(best_rank.token_match_start_index,
                            best_rank.token_match_end_index)]
            
                setattr(token._, self.pymusas_tags_token_attr, tags)
                setattr(token._, self.pymusas_mwe_indexes_attr, indexes)
        except Exception as e:
            error_handler(self.name, self, [doc], e)
        
        return doc

    def to_bytes(self, *, exclude: Iterable[str] = SimpleFrozenList()) -> bytes:
        '''
        Serialises the tagger to a bytestring.

        # Parameters

        exclude : `Iterable[str]`, optional (default = `SimpleFrozenList()`)
            This currently does not do anything, please ignore it.
        
        # Returns

        `bytes`

        # Examples

        ``` python
        >>> from pymusas.rankers.lexicon_entry import ContextualRuleBasedRanker
        >>> from pymusas.taggers.rules.single_word import SingleWordRule
        >>> from pymusas.spacy_api.taggers.rule_based import RuleBasedTagger
        >>> rules = [SingleWordRule({'example|noun': ['Z1']}, {})]
        >>> ranker = ContextualRuleBasedRanker(*ContextualRuleBasedRanker.get_construction_arguments(rules))
        >>> tagger = RuleBasedTagger()
        >>> tagger.initialize(rules=rules, ranker=ranker)
        >>> tagger_bytes = tagger.to_bytes()

        ```
        '''
        if not self._validated:
            self._validate()
        ranker = cast(LexiconEntryRanker, self.ranker)
        rules = cast(List[Rule], self.rules)

        default_punctuation_tags = list(self.default_punctuation_tags)
        default_number_tags = list(self.default_number_tags)
        
        serialise = {}
        serialise["rules"] = Rule.serialise_object_list_to_bytes(rules)
        serialise["ranker"] = LexiconEntryRanker.serialise_object_to_bytes(ranker)
        serialise["default_punctuation_tags"] = srsly.msgpack_dumps(default_punctuation_tags)
        serialise["default_number_tags"] = srsly.msgpack_dumps(default_number_tags)
        
        return cast(bytes, srsly.msgpack_dumps(serialise))

    def from_bytes(self, bytes_data: bytes, *,
                   exclude: Iterable[str] = SimpleFrozenList()
                   ) -> "RuleBasedTagger":
        '''
        Loads the tagger from the given bytestring in place and returns it.

        # Parameters

        bytes_data : `bytes`
            The bytestring to load.
        exclude : `Iterable[str]`, optional (default = `SimpleFrozenList()`)
            This currently does not do anything, please ignore it.

        # Returns

        :class:`RuleBasedTagger`

        # Examples
        
        ``` python
        >>> from pymusas.rankers.lexicon_entry import ContextualRuleBasedRanker
        >>> from pymusas.taggers.rules.single_word import SingleWordRule
        >>> from pymusas.spacy_api.taggers.rule_based import RuleBasedTagger
        >>> rules = [SingleWordRule({'example|noun': ['Z1']}, {})]
        >>> ranker = ContextualRuleBasedRanker(*ContextualRuleBasedRanker.get_construction_arguments(rules))
        >>> tagger = RuleBasedTagger()
        >>> tagger.initialize(rules=rules, ranker=ranker)
        >>> # Create a new tagger, tagger 2
        >>> tagger_2 = RuleBasedTagger()
        >>> # Show that it is not the same as the original tagger
        >>> assert tagger_2.rules != rules
        >>> # Tagger 2 will now load in the data from the original tagger
        >>> _ = tagger_2.from_bytes(tagger.to_bytes())
        >>> assert tagger_2.rules == rules
        >>> assert tagger_2.ranker == ranker

        ```
        '''
        serialise_data = srsly.msgpack_loads(bytes_data)

        self.rules = cast(List[Rule],
                          Rule.serialise_object_list_from_bytes(serialise_data['rules']))

        self.ranker \
            = cast(LexiconEntryRanker,
                   LexiconEntryRanker.serialise_object_from_bytes(serialise_data['ranker']))

        default_punctuation_tags = srsly.msgpack_loads(serialise_data["default_punctuation_tags"])
        self.default_punctuation_tags = set(default_punctuation_tags)

        default_number_tags = srsly.msgpack_loads(serialise_data["default_number_tags"])
        self.default_number_tags = set(default_number_tags)
        
        self._validate()
        return self
    
    def to_disk(self, path: Union[str, Path], *,
                exclude: Iterable[str] = SimpleFrozenList()
                ) -> None:
        '''
        Serialises the tagger to the given `path`.

        # Parameters

        path : `Union[str, Path]`
            Path to a direcotry. Path may be either string or `Path`-like
            object. If the directory does not exist it attempts to create a
            directory at the given `path`.
        
        exclude : `Iterable[str]`, optional (default = `SimpleFrozenList()`)
            This currently does not do anything, please ignore it.

        # Returns

        `None`

        # Examples

        ```python
        >>> from pathlib import Path
        >>> from tempfile import TemporaryDirectory
        >>> from pymusas.rankers.lexicon_entry import ContextualRuleBasedRanker
        >>> from pymusas.taggers.rules.single_word import SingleWordRule
        >>> from pymusas.spacy_api.taggers.rule_based import RuleBasedTagger
        >>> rules = [SingleWordRule({'example|noun': ['Z1']}, {})]
        >>> ranker = ContextualRuleBasedRanker(*ContextualRuleBasedRanker.get_construction_arguments(rules))
        >>> tagger = RuleBasedTagger()
        >>> tagger.initialize(rules=rules, ranker=ranker)
        >>> with TemporaryDirectory() as temp_dir:
        ...     _ = tagger.to_disk(temp_dir)
        ...

        ```
        '''
        if not self._validated:
            self._validate()
        ranker = cast(LexiconEntryRanker, self.ranker)
        rules = cast(List[Rule], self.rules)

        component_folder = ensure_path(path)
        component_folder.mkdir(exist_ok=True)

        ranker_path = Path(component_folder, 'ranker.bin')
        srsly.write_msgpack(ranker_path,
                            LexiconEntryRanker.serialise_object_to_bytes(ranker))

        rules_path = Path(component_folder, 'rules.bin')
        srsly.write_msgpack(rules_path,
                            Rule.serialise_object_list_to_bytes(rules))

        default_punctuation_tags_path = Path(component_folder,
                                             'default_punctuation_tags.bin')
        default_punctuation_tags = list(self.default_punctuation_tags)
        srsly.write_msgpack(default_punctuation_tags_path, default_punctuation_tags)
        
        default_number_tags_path = Path(component_folder, 'default_number_tags.bin')
        default_number_tags = list(self.default_number_tags)
        srsly.write_msgpack(default_number_tags_path, default_number_tags)

    def from_disk(self, path: Union[str, Path], *,
                  exclude: Iterable[str] = SimpleFrozenList()
                  ) -> "RuleBasedTagger":
        '''
        Loads the tagger from the given `path` in place and returns it.

        # Parameters

        path : `Union[str, Path]`
            Path to an existing direcotry. Path may be either string or
            `Path`-like object.
        
        exclude : `Iterable[str]`, optional (default = `SimpleFrozenList()`)
            This currently does not do anything, please ignore it.

        # Returns

        :class:`RuleBasedTagger`

        # Examples

        ```python
        >>> from pathlib import Path
        >>> from tempfile import TemporaryDirectory
        >>> from pymusas.rankers.lexicon_entry import ContextualRuleBasedRanker
        >>> from pymusas.taggers.rules.single_word import SingleWordRule
        >>> from pymusas.spacy_api.taggers.rule_based import RuleBasedTagger
        >>> rules = [SingleWordRule({'example|noun': ['Z1']}, {})]
        >>> ranker = ContextualRuleBasedRanker(*ContextualRuleBasedRanker.get_construction_arguments(rules))
        >>> tagger = RuleBasedTagger()
        >>> tagger.initialize(rules=rules, ranker=ranker)
        >>> # Create an empty second tagger
        >>> tagger_2 = RuleBasedTagger()
        >>> assert tagger_2.rules is None
        >>> with TemporaryDirectory() as temp_dir:
        ...     _ = tagger.to_disk(temp_dir)
        ...     _ = tagger_2.from_disk(temp_dir)
        ...
        >>> assert tagger_2.rules is not None
        >>> assert tagger_2.rules == tagger.rules

        ```

        '''
        component_folder = ensure_path(path)
        
        ranker_path = Path(component_folder, 'ranker.bin')
        serialised_ranker = srsly.read_msgpack(ranker_path)
        ranker = LexiconEntryRanker.serialise_object_from_bytes(serialised_ranker)
        self.ranker = cast(LexiconEntryRanker, ranker)

        rules_path = Path(component_folder, 'rules.bin')
        serialised_rules = srsly.read_msgpack(rules_path)
        rules = Rule.serialise_object_list_from_bytes(serialised_rules)
        self.rules = cast(List[Rule], rules)

        default_punctuation_tags_path = Path(component_folder,
                                             'default_punctuation_tags.bin')
        default_punctuation_tags = srsly.read_msgpack(default_punctuation_tags_path)
        self.default_punctuation_tags = set(default_punctuation_tags)
        
        default_number_tags_path = Path(component_folder, 'default_number_tags.bin')
        default_number_tags = srsly.read_msgpack(default_number_tags_path)
        self.default_number_tags = set(default_number_tags)

        self._validate()
        return self
        
    @property
    def pymusas_tags_token_attr(self) -> str:
        return self._pymusas_tags_token_attr

    @property
    def pymusas_mwe_indexes_attr(self) -> str:
        return self._pymusas_mwe_indexes_attr

    @property
    def pos_attribute(self) -> str:
        return self._pos_attribute

    @property
    def lemma_attribute(self) -> str:
        return self._lemma_attribute


@Language.factory(RuleBasedTagger.COMPONENT_NAME, requires=['token.pos', 'token.lemma'],
                  assigns=['token._.pymusas_tags', 'token._.pymusas_mwe_indexes'],
                  default_config={'pymusas_tags_token_attr': 'pymusas_tags',
                                  'pymusas_mwe_indexes_attr': 'pymusas_mwe_indexes',
                                  'pos_attribute': 'pos_',
                                  'lemma_attribute': 'lemma_'})
def make_usas_rule_based_tagger(nlp: Language, name: str,
                                pymusas_tags_token_attr: str,
                                pymusas_mwe_indexes_attr: str,
                                pos_attribute: str,
                                lemma_attribute: str
                                ) -> RuleBasedTagger:
    return RuleBasedTagger(name, pymusas_tags_token_attr,
                           pymusas_mwe_indexes_attr,
                           pos_attribute, lemma_attribute)



from abc import abstractmethod
from typing import List

from pymusas.base import Serialise
from pymusas.rankers.ranking_meta_data import RankingMetaData


class Rule(Serialise):
    '''
    An **abstract class** that defines the basic method, `__call__`, that is
    required for all :class:`Rule`s.

    A Rule when called, `__call__`, creates a `List` of rules matches for each
    token, whereby each rule matched is defined by the
    :class:`pymusas.rankers.ranking_meta_data.RankingMetaData` object. These
    rules matches per token can then be, optionally, combined with other rule
    matches per token from other :class:`Rule` classes to then be ranked by a
    :class:`pymusas.rankers.lexicon_entry.LexiconEntryRanker`.
    '''

    @abstractmethod
    def __call__(self, tokens: List[str], lemmas: List[str],
                 pos_tags: List[str]) -> List[List[RankingMetaData]]:
        '''
        For each token it returns a `List` of rules matches defined by the
        :class:`pymusas.rankers.ranking_meta_data.RankingMetaData` object.

        Each `List` of `tokens`, `lemmas`, and `pos_tags` are assumed to be of
        equal length.

        # Parameters

        tokens : `List[str]`
            The tokens that are within the text.
        lemmas : `List[str]`
            The lemmas of the tokens.
        pos_tags : `List[str]`
            The Part Of Speech tags of the tokens.

        # Returns

        `List[List[RankingMetaData]]`
        '''
        ...  # pragma: no cover

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        ...  # pragma: no cover



'''
spaCy registered functions for reading in a
:class:`pymusas.lexicon_collection.LexiconCollection` or
:class:`pymusas.lexicon_collection.MWELexiconCollection` from a TSV file.
'''

from os import PathLike
from typing import Dict, List, Union

import spacy

from ..lexicon_collection import LexiconCollection, MWELexiconCollection


@spacy.util.registry.misc('pymusas.LexiconCollection.from_tsv')
def lexicon_collection_from_tsv(tsv_file_path: Union[PathLike, str],
                                include_pos: bool = True
                                ) -> Dict[str, List[str]]:
    '''
    `pymusas.LexiconCollection.from_tsv` is a registered function under the
    `@misc` function register. Given a `tsv_file_path` it will return a
    dictionary object that can be used to create a
    :class:`pymusas.lexicon_collection.LexiconCollection`.

    # Parameters

    tsv_file_path: `Union[PathLike, str]`
        A file path or URL to a TSV file that contains at least two
        fields, with an optional third, with the following headings:
        
        1. `lemma`,
        2. `semantic_tags`
        3. `pos` (Optional)
        
        All other fields will be ignored.
    include_pos: `bool`, optional (default = `True`)
        Whether to include the POS information, if the information is avaliable,
        or not. See :func:`add_lexicon_entry` for more information on this
        parameter.

    # Returns
    
    `Dict[str, List[str]]`
    '''
    return LexiconCollection.from_tsv(tsv_file_path, include_pos)


@spacy.util.registry.misc('pymusas.MWELexiconCollection.from_tsv')
def mwe_lexicon_collection_from_tsv(tsv_file_path: Union[PathLike, str]
                                    ) -> Dict[str, List[str]]:
    '''
    `pymusas.MWELexiconCollection.from_tsv` is a registered function under the
    `@misc` function register. Given a `tsv_file_path` it will return a
    dictionary object that can be used to create a
    :class:`pymusas.lexicon_collection.MWELexiconCollection`.

    # Parameters

    tsv_file_path: `Union[PathLike, str]`
        A file path or URL to a TSV file that contains at least these two
        fields:
        
        1. `mwe_template`,
        2. `semantic_tags`
        
        All other fields will be ignored.

    # Returns
    
    `Dict[str, List[str]]`
    '''
    return MWELexiconCollection.from_tsv(tsv_file_path)



from typing import Iterable, Set, Tuple


def token_pos_tags_in_lexicon_entry(lexicon_entry: str
                                    ) -> Iterable[Tuple[str, str]]:
    '''
    Yields the token and associated POS tag in the given `lexicon_entry`.

    # Parameters

    lexicon_entry : `str`
        Either a Multi Word Expression template or single word lexicon entry,
        which is a sequence of words/tokens and Part Of Speech (POS) tags
        joined together by an underscore and separated by a single whitespace,
        e.g. `word1_POS1 word2_POS2 word3_POS3`. For a single word lexicon it
        would be `word1_POS1`.

    # Returns

    `Iterable[Tuple[str, str]]`

    # Raises

    `ValueError`
        If the lexicon entry when split on whitespace and then split by `_`
        does not create a `Iterable[Tuple[str, str]]` whereby the tuple contains
        the `token text` and it's associated `POS tag`.

    # Examples
    ``` python
    >>> from pymusas.utils import token_pos_tags_in_lexicon_entry
    >>> mwe_template = 'East_noun London_noun is_det great_adj'
    >>> assert ([('East', 'noun'), ('London', 'noun'), ('is', 'det'), ('great', 'adj')]
    ...         == list(token_pos_tags_in_lexicon_entry(mwe_template)))
    >>> single_word_lexicon = 'East_noun'
    >>> assert ([('East', 'noun')]
    ...         == list(token_pos_tags_in_lexicon_entry(single_word_lexicon)))

    ```
    '''
    split_error_msg = ('The lexicon entry when split on whitespace and then by'
                       ' `_` does not create a `List[Tuple[str, str]]`. '
                       f'lexicon entry: {lexicon_entry}')
    for token_pos in lexicon_entry.split():
        token_pos_split = token_pos.split('_')
        if len(token_pos_split) != 2:
            raise ValueError(split_error_msg)
        token, pos_tag = token_pos_split
        yield token, pos_tag


def unique_pos_tags_in_lexicon_entry(lexicon_entry: str) -> Set[str]:
    '''
    Returns the unique POS tag values in the given `lexicon_entry`.

    # Parameters

    lexicon_entry : `str`
        Either a Multi Word Expression template or single word lexicon entry,
        which is a sequence of words/tokens and Part Of Speech (POS) tags
        joined together by an underscore and separated by a single whitespace,
        e.g. `word1_POS1 word2_POS2 word3_POS3`. For a single word lexicon it
        would be `word1_POS1`.

    # Returns

    `Set[str]`

    # Raises

    `ValueError`
        If the lexicon entry when split on whitespace and then split by `_`
        does not create a `List[Tuple[str, str]]` whereby the tuple contains
        the `token text` and it's associated `POS tag`.

    # Examples
    ``` python
    >>> from pymusas.utils import unique_pos_tags_in_lexicon_entry
    >>> mwe_template = 'East_noun London_noun is_det great_adj'
    >>> assert ({'noun', 'adj', 'det'}
    ...         == unique_pos_tags_in_lexicon_entry(mwe_template))
    >>> single_word_lexicon = 'East_noun'
    >>> assert {'noun'} == unique_pos_tags_in_lexicon_entry(single_word_lexicon)

    ```
    '''
    unique_pos_tags: Set[str] = set()
    for _, pos_tag in token_pos_tags_in_lexicon_entry(lexicon_entry):
        unique_pos_tags.add(pos_tag)
    return unique_pos_tags



from typing import Dict, List, Optional, cast

import srsly

from pymusas.lexicon_collection import LexiconCollection, LexiconType
from pymusas.rankers.lexical_match import LexicalMatch
from pymusas.rankers.ranking_meta_data import RankingMetaData
from pymusas.taggers.rules.rule import Rule


class SingleWordRule(Rule):
    '''
    A single word rule match, is a rule that matches on single word lexicon
    entries. Entires can be matched on:
    
    1. Token and the token's Part Of Speech (POS) tag, e.g. `driving|adj`
    2. Lemma and the lemma's POS tag, e.g. `drive|adj`
    3. Token, e.g. `driving`
    4. Lemma, e.g. `drive`

    In all cases matches are found based on the original token/lemma and lower
    cased versions of the token/lemma. These matches are found through searching
    the `lexicon_collection` and `lemma_lexicon_collection` attributes.


    # Parameters

    lexicon_collection : `Dict[str, List[str]]`
        The data to create `lexicon_collection` instance attribute. A
        Dictionary where the keys are a combination of
        lemma/token and POS in the following format: `{lemma}|{POS}` and the
        values are a list of associated semantic tags.
    lemma_lexicon_collection : `Dict[str, List[str]]`
        The data to create `lemma_lexicon_collection` instance attribute. A
        Dictionary where the keys are either just a lemma/token
        in the following format: `{lemma}` and the
        values are a list of associated semantic tags.
    pos_mapper : `Dict[str, List[str]]`, optional (default = `None`)
        If not `None`, maps from the given token's POS tagset to the desired
        POS tagset, whereby the mapping is a `List` of tags, at the moment there
        is no preference order in this list of POS tags. The POS mapping is
        useful in situtation whereby the token's POS tagset is different to
        those used in the lexicons. **Note** the longer the `List[str]` for
        each POS mapping the slower the tagger, a one to one mapping will have
        no speed impact on the tagger. A selection of POS mappers can be found in
        :mod:`pymusas.pos_mapper`.

    # Instance Attributes

    lexicon_collection : `pymusas.lexicon_collection.LexiconCollection`
        A :class:`pymusas.lexicon_collection.LexiconCollection` instance that
        has been initialised using the `lexicon_collection` parameter.
    lemma_lexicon_collection : `pymusas.lexicon_collection.LexiconCollection`
        A :class:`pymusas.lexicon_collection.LexiconCollection` instance that
        has been initialised using the `lemma_lexicon_collection` parameter.
    pos_mapper : `Dict[str, List[str]]`, optional (default = `None`)
        The given `pos_mapper`.
    '''
    def __init__(self, lexicon_collection: Dict[str, List[str]],
                 lemma_lexicon_collection: Dict[str, List[str]],
                 pos_mapper: Optional[Dict[str, List[str]]] = None):

        self.lexicon_collection = LexiconCollection(lexicon_collection)
        self.lemma_lexicon_collection = LexiconCollection(lemma_lexicon_collection)
        self.pos_mapper = pos_mapper

    def __call__(self, tokens: List[str], lemmas: List[str], pos_tags: List[str]
                 ) -> List[List[RankingMetaData]]:
        '''
        Given the tokens, lemmas, and POS tags for each word in a text,
        it returns for each token a `List` of rules matches defined by
        the :class:`pymusas.rankers.ranking_meta_data.RankingMetaData`
        object based on the rule matches stated in the class docstring above.
        
        # Parameters

        tokens : `List[str]`
            The tokens that are within the text.
        lemmas : `List[str]`
            The lemmas of the tokens.
        pos_tags : `List[str]`
            The Part Of Speech tags of the tokens.
        
        # Returns

        `List[List[RankingMetaData]]`
        '''

        def find_match_and_add_to_ranking_data(lexicon_entry: str,
                                               exclude_pos_information: bool,
                                               lexical_match: LexicalMatch,
                                               start_index: int, end_index: int,
                                               token_ranking_meta_data: List[List[RankingMetaData]]
                                               ) -> None:
            collection = self.lexicon_collection
            if exclude_pos_information:
                collection = self.lemma_lexicon_collection
            if lexicon_entry in collection:
                semantic_tags = tuple(collection[lexicon_entry])
                ranking_data = RankingMetaData(LexiconType.SINGLE_NON_SPECIAL,
                                               1, 0, exclude_pos_information,
                                               lexical_match,
                                               start_index, end_index,
                                               lexicon_entry, semantic_tags)
                token_ranking_meta_data[start_index].append(ranking_data)

        token_ranking_meta_data: List[List[RankingMetaData]] \
            = [[] for _ in range(len(tokens))]
        
        index = 0

        for token, lemma, initial_pos in zip(tokens, lemmas, pos_tags):

            token_lower = token.lower()
            lemma_lower = lemma.lower()

            start_index = index
            end_index = start_index + 1
            
            pos_tags = [initial_pos]
            if self.pos_mapper is not None:
                pos_tags = self.pos_mapper.get(initial_pos, [])
            
            # All of these use POS information
            for pos in pos_tags:
                token_pos = f'{token}|{pos}'
                find_match_and_add_to_ranking_data(token_pos, False,
                                                   LexicalMatch.TOKEN,
                                                   start_index, end_index,
                                                   token_ranking_meta_data)
                
                lemma_pos = f'{lemma}|{pos}'
                find_match_and_add_to_ranking_data(lemma_pos, False,
                                                   LexicalMatch.LEMMA,
                                                   start_index, end_index,
                                                   token_ranking_meta_data)
                
                token_lower_pos = f'{token_lower}|{pos}'
                find_match_and_add_to_ranking_data(token_lower_pos, False,
                                                   LexicalMatch.TOKEN_LOWER,
                                                   start_index, end_index,
                                                   token_ranking_meta_data)
                
                lemma_lower_pos = f'{lemma_lower}|{pos}'
                find_match_and_add_to_ranking_data(lemma_lower_pos, False,
                                                   LexicalMatch.LEMMA_LOWER,
                                                   start_index, end_index,
                                                   token_ranking_meta_data)
                
            # All of these do not use POS information
            lexical_value_type = [(token, LexicalMatch.TOKEN),
                                  (lemma, LexicalMatch.LEMMA),
                                  (token_lower, LexicalMatch.TOKEN_LOWER),
                                  (lemma_lower, LexicalMatch.LEMMA_LOWER)]
            for lexical_value, lexical_type in lexical_value_type:
                find_match_and_add_to_ranking_data(lexical_value, True, lexical_type,
                                                   start_index, end_index,
                                                   token_ranking_meta_data)

            index += 1

        return token_ranking_meta_data

    def to_bytes(self) -> bytes:
        '''
        Serialises the :class:`SingleWordRule` to a bytestring.

        # Returns

        `bytes`
        '''
        serialise = {}
        serialise['lexicon_collection'] = self.lexicon_collection.to_bytes()
        serialise['lemma_lexicon_collection'] \
            = self.lemma_lexicon_collection.to_bytes()
        serialise['pos_mapper'] = srsly.msgpack_dumps(self.pos_mapper)
        return cast(bytes, srsly.msgpack_dumps(serialise))

    @staticmethod
    def from_bytes(bytes_data: bytes) -> "SingleWordRule":
        '''
        Loads :class:`SingleWordRule` from the given bytestring and returns it.

        # Parameters

        bytes_data : `bytes`
            The bytestring to load.
        
        # Returns

        :class:`SingleWordRule`
        '''
        serialise_data = srsly.msgpack_loads(bytes_data)
        lexicon_collection \
            = LexiconCollection.from_bytes(serialise_data['lexicon_collection'])
        lemma_lexicon_collection \
            = LexiconCollection.from_bytes(serialise_data['lemma_lexicon_collection'])
        pos_mapper = srsly.msgpack_loads(serialise_data['pos_mapper'])
        
        single_word_rule = SingleWordRule({}, {}, None)
        single_word_rule.lexicon_collection = lexicon_collection
        single_word_rule.lemma_lexicon_collection = lemma_lexicon_collection
        single_word_rule.pos_mapper = pos_mapper
        return single_word_rule

    def __eq__(self, other: object) -> bool:
        '''
        Given another object to compare too it will return `True` if the other
        object is the same class and initialised using with the same argument
        values.

        # Parameters

        other : `object`
            The object to compare too.
        
        # Returns

        `True`
        '''
        if not isinstance(other, SingleWordRule):
            return False
        
        if self.pos_mapper != other.pos_mapper:
            return False
        
        if self.lexicon_collection != other.lexicon_collection:
            return False

        if self.lemma_lexicon_collection != other.lemma_lexicon_collection:
            return False
        
        return True



'''
This module has various attributes, of which the most important of these
are listed below:

# Attributes

PYMUSAS_CACHE_HOME: `str`
       The directory that by default we store any downloaded data too. This
       attribute by default is set to `~/.cache/pymusas`. This attribute can be
       set through the `PYMUSAS_HOME` environment variable.

The creation of the `PYMUSAS_CACHE_HOME` attribute and how to set a default value
for it came from the [HuggingFace Datasets codebase
(reference to their code)](https://github.com/huggingface/datasets/blob/d488db2f64f312f88f72bbc57a09b7eddb329182/src/datasets/config.py#L130).
'''
import os


DEFAULT_XDG_CACHE_HOME: str = os.path.join(os.path.expanduser('~'), '.cache')
XDG_CACHE_HOME: str = os.getenv("XDG_CACHE_HOME", DEFAULT_XDG_CACHE_HOME)
DEFAULT_PYMUSAS_CACHE_HOME: str = os.path.join(XDG_CACHE_HOME, "pymusas")
PYMUSAS_CACHE_HOME: str = os.path.expanduser(os.getenv("PYMUSAS_HOME", DEFAULT_PYMUSAS_CACHE_HOME))

LANG_LEXICON_RESOUCRE_MAPPER = {
    'fr': {'lexicon': 'https://raw.githubusercontent.com/UCREL/Multilingual-USAS/master/French/semantic_lexicon_fr.usas',
           'lexicon_lemma': 'https://raw.githubusercontent.com/UCREL/Multilingual-USAS/master/French/semantic_lexicon_fr.usas'},
    'nl': {'lexicon': 'https://raw.githubusercontent.com/UCREL/Multilingual-USAS/master/Dutch/semantic_lexicon_dut.usas',
           'lexicon_lemma': 'https://raw.githubusercontent.com/UCREL/Multilingual-USAS/master/Dutch/semantic_lexicon_dut.usas'}
}



from typing import List, Optional, Set, Tuple

from pymusas.rankers.lexicon_entry import LexiconEntryRanker, RankingMetaData
from pymusas.taggers.rules.rule import Rule


class RuleBasedTagger():
    '''
    The tagger when called, through :func:`__call__`, and given a sequence of
    tokens and their associated lingustic data (lemma, Part Of Speech (POS))
    will apply one or more :class:`pymusas.taggers.rules.rule.Rule`s
    to create a list of possible candidate tags for each token in the sequence.
    Each candidate, represented as a
    :class:`pymusas.rankers.ranking_meta_data.RankingMetaData` object, for each
    token is then Ranked using a
    :class:`pymusas.rankers.lexicon_entry.LexiconEntryRanker` ranker. The best
    candidate and it's associated tag(s) for each token are then returned along
    with a `List` of token indexes indicating if the token is part of a Multi
    Word Expression (MWE).

    If we cannot tag a token then the following process will happen:
    1. If the token's POS tag is in `default_punctuation_tags` then it will assign the
    tag `PUNCT`.
    2. If the token's POS tag is in `default_number_tags` then it will assign the tag
    `N1`.
    3. Assign the default tag `Z99`.

    # Parameters

    rules : `List[pymusas.taggers.rules.rule.Rule]`
        A list of rules to apply to the sequence of tokens in the
        :func:`__call__`. The output from each rule is concatendated and given
        to the `ranker`.
    ranker : `pymusas.rankers.lexicon_entry.LexiconEntryRanker`
        A ranker to rank the output from all of the `rules`.
    default_punctuation_tags : `Set[str]`, optional (default = `None`)
        The POS tags that represent punctuation. If `None` then we will use
        the `Set`: `set(['punc'])`.
    default_number_tags : `Set[str]`, optional (default = `None`)
        The POS tags that represent numbers. If `None` then we will use
        the `Set`: `set(['num'])`.
    
    # Instance Attributes

    rules : `List[pymusas.taggers.rules.rule.Rule]`
        The given `rules`.
    ranker : `pymusas.rankers.lexicon_entry.LexiconEntryRanker`
        The given `ranker`.
    default_punctuation_tags : `Set[str]`
        The given `default_punctuation_tags`
    default_number_tags : `Set[str]`
        The given `default_number_tags`

    # Examples
    ``` python
    >>> from pymusas.lexicon_collection import LexiconCollection
    >>> from pymusas.taggers.rule_based import RuleBasedTagger
    >>> from pymusas.taggers.rules.single_word import SingleWordRule
    >>> from pymusas.rankers.lexicon_entry import ContextualRuleBasedRanker
    >>> from pymusas.pos_mapper import BASIC_CORCENCC_TO_USAS_CORE
    >>> welsh_lexicon_url = 'https://raw.githubusercontent.com/apmoore1/Multilingual-USAS/master/Welsh/semantic_lexicon_cy.tsv'
    >>> lexicon_lookup = LexiconCollection.from_tsv(welsh_lexicon_url, include_pos=True)
    >>> lemma_lexicon_lookup = LexiconCollection.from_tsv(welsh_lexicon_url, include_pos=False)
    >>> single_word_rule = SingleWordRule(lexicon_lookup, lemma_lexicon_lookup,
    ...                                   BASIC_CORCENCC_TO_USAS_CORE)
    >>> ranker = ContextualRuleBasedRanker(1, 0)
    >>> tagger = RuleBasedTagger([single_word_rule], ranker)

    ```
    '''

    def __init__(self, rules: List[Rule], ranker: LexiconEntryRanker,
                 default_punctuation_tags: Optional[Set[str]] = None,
                 default_number_tags: Optional[Set[str]] = None) -> None:

        self.rules = rules
        self.ranker = ranker

        self.default_punctuation_tags = set(['punc'])
        if default_punctuation_tags is not None:
            self.default_punctuation_tags = default_punctuation_tags
        
        self.default_number_tags = set(['num'])
        if default_number_tags is not None:
            self.default_number_tags = default_number_tags

    def __call__(self, tokens: List[str], lemmas: List[str],
                 pos_tags: List[str]) -> List[Tuple[List[str],
                                                    List[Tuple[int, int]]
                                                    ]]:
        '''
        Given a `List` of tokens, their associated lemmas and
        Part Of Speech (POS) tags it returns for each token:
        
        1. A `List` of tags. The first tag in the `List` of tags is the most likely tag.
        2. A `List` of `Tuples` whereby each `Tuple` indicates the start and end
        token index of the associated Multi Word Expression (MWE). If the `List` contains
        more than one `Tuple` then the MWE is discontinuous. For single word
        expressions the `List` will only contain 1 `Tuple` which will be
        (token_start_index, token_start_index + 1).
    
        All the generated tags and MWEs are based on the rules and ranker given
        to this model.

        **NOTE** this tagger has been designed to be flexible with the amount of
        resources avaliable, if you do not have POS or lemma information assign
        them a `List` of empty strings.
        
        # Parameters

        tokens : `List[str]`
            A List of full text form of the tokens to be tagged.
        lemmas : `List[str]`
            The List of lemma/base form of the tokens to be tagged.
        pos_tags : `List[str]`
            The List of POS tags of the tokens to be tagged.
        
        # Returns

        `List[Tuple[List[str], List[Tuple[int, int]]]]`

        # Raises

        `ValueError`
            If the length of the `tokens`, `lemmas`, and `pos_tags` are not of
            the same legnth.
        '''

        tokens_length = len(tokens)
        pos_tags_length = len(pos_tags)
        lemmas_length = len(lemmas)
        length_error_msg = ('The `tokens`, `lemmas`, or `pos_tags` are not '
                            'of the the same length, their lengths respectively:'
                            f' {tokens_length}, {pos_tags_length}, {lemmas_length}')
        if (tokens_length != pos_tags_length) or (tokens_length != lemmas_length):
            raise ValueError(length_error_msg)
            
        token_ranking_meta_data: List[List[RankingMetaData]] \
            = [[] for _ in range(len(tokens))]
        for rule in self.rules:
            rule_ranking_meta_data = rule(tokens, lemmas, pos_tags)
            for token_index, ranking_meta_data in enumerate(rule_ranking_meta_data):
                token_ranking_meta_data[token_index].extend(ranking_meta_data)

        token_ranks, token_best_rank = self.ranker(token_ranking_meta_data)
        
        tags_indexes: List[Tuple[List[str], List[Tuple[int, int]]]] = []
        for token_index, best_rank in enumerate(token_best_rank):
            if best_rank is None:
                pos_tag = pos_tags[token_index]
                if pos_tag in self.default_punctuation_tags:
                    tags_indexes.append((['PUNCT'],
                                         [(token_index, token_index + 1)]))
                elif pos_tag in self.default_number_tags:
                    tags_indexes.append((['N1'],
                                         [(token_index, token_index + 1)]))
                else:
                    tags_indexes.append((['Z99'],
                                         [(token_index, token_index + 1)]))
                continue
            tags = list(best_rank.semantic_tags)
            indexes = [(best_rank.token_match_start_index,
                        best_rank.token_match_end_index)]
            tags_indexes.append((tags, indexes))
        
        return tags_indexes



from typing import Any, Iterator, Sequence, Tuple


def n_gram_indexes(sequence: Sequence[Any], min_n: int, max_n: int
                   ) -> Iterator[Tuple[int, int]]:
    '''
    Returns n-grams as indexes of the `sequence`,
    in the range from `max_n` to `min_n`, in
    order of largest n-grams first. If you only want one n-gram size then set
    `min_n` equal to `max_n`, for example to get bi-grams indexes set both
    `min_n` and `max_n` to `2`.

    # Parameters

    sequence : `Sequence[Any]`
        The sequence to generate n-gram indexes from.
    min_n : `int`
        Minimum size n-gram. Has to be greater than `0`.
    max_n : `int`
        Maximim size n-gram. This has to be equal to or greater than `min_n`.
        If this is greater than the length of the `sequence` then it is set to
        length of the `sequence`.

    # Returns

    `Iterator[Tuple[int, int]]`

    # Raises

    `ValueError`
        If `min_n` is less than `1` or `max_n` is less than `min_n`.

    # Examples

    ``` python
    >>> from pymusas.taggers.rules.util import n_gram_indexes
    >>> tokens = ['hello', 'how', 'are', 'you', ',']
    >>> token_n_gram_indexes = n_gram_indexes(tokens, 2, 3)
    >>> expected_n_grams_indexes = [(0, 3), (1, 4), (2, 5), (0, 2), (1, 3), (2, 4), (3, 5)]
    >>> assert expected_n_grams_indexes == list(token_n_gram_indexes)

    ```
    '''
    if min_n < 1:
        raise ValueError(f'`min_n` has to be greater than 0. Value given {min_n}')
    if min_n > max_n:
        raise ValueError(f'`max_n` ({max_n}) cannot be less than `min_n` ({min_n})')
    
    sequence_size = len(sequence)
    max_n = sequence_size if max_n > sequence_size else max_n
    
    for n_gram_size in range(max_n, min_n - 1, -1):
        for sequence_index in range(sequence_size - n_gram_size + 1):
            last_n_gram_index = sequence_index + n_gram_size
            yield (sequence_index, last_n_gram_index)


def n_grams(sequence: Sequence[Any], min_n: int, max_n: int) -> Iterator[Sequence[Any]]:
    '''
    Returns n-grams, in the range from `max_n` to `min_n`, of the `sequence` in
    order of largest n-grams first. If you only want one n-gram size then set
    `min_n` equal to `max_n`, for example to get bi-grams set both `min_n` and
    `max_n` to `2`.

    # Parameters

    sequence : `Sequence[Any]`
        The sequence to generate n-grams from.
    min_n : `int`
        Minimum size n-gram. Has to be greater than `0`.
    max_n : `int`
        Maximim size n-gram. This has to be equal to or greater than `min_n`.
        If this is greater than the length of the `sequence` then it is set to
        length of the `sequence`.

    # Returns

    `Iterator[Sequence[Any]]`

    # Raises

    `ValueError`
        If `min_n` is less than `1` or `max_n` is less than `min_n`.

    # Examples

    ``` python
    >>> from pymusas.taggers.rules.util import n_grams
    >>> tokens = ['hello', 'how', 'are', 'you', ',']
    >>> token_n_grams = n_grams(tokens, 2, 3)
    >>> expected_n_grams = [['hello', 'how', 'are'], ['how', 'are', 'you'], ['are', 'you', ','],
    ...                     ['hello', 'how'], ['how', 'are'], ['are', 'you'], ['you', ',']]
    >>> assert expected_n_grams == list(token_n_grams)

    ```
    '''
    for n_gram_index in n_gram_indexes(sequence, min_n, max_n):
        start_index, end_index = n_gram_index
        yield sequence[start_index: end_index]



from hashlib import sha256
import os
from pathlib import Path
from typing import Union

import requests
from requests.adapters import HTTPAdapter
from tqdm import tqdm
from urllib3.util.retry import Retry

from . import config


def ensure_path(path: Union[str, Path]) -> Path:
    """
    Ensure string is converted to a Path.
    
    This is a more restrictive version of spaCy's [ensure_path](https://github.com/explosion/spaCy/blob/ac05de2c6c708e33ebad6c901e674e1e8bdc0688/spacy/util.py#L358)
    
    # Parameters

    path : `Union[str, Path]`
        If string, it's converted to Path.
    
    # Returns

    `Path`
    """
    if isinstance(path, str):
        return Path(path)
    return path


def _session_with_backoff() -> requests.Session:
    """
    We ran into an issue where http requests to s3 were timing out,
    possibly because we were making too many requests too quickly.
    This helper function returns a requests session that has retry-with-backoff
    built in. See this [Stack Overflow post.](https://stackoverflow.com/questions/23267409/how-to-implement-retry-mechanism-into-python-requests-library>.)
    
    Code reference [AllenNLP](https://github.com/allenai/allennlp/blob/e5d332a592a8624e1f4ee7a9a7d30a90991db83c/allennlp/common/file_utils.py#L510)

    # Returns

    `requests.Session`
    """
    session = requests.Session()
    retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504])
    session.mount("http://", HTTPAdapter(max_retries=retries))
    session.mount("https://", HTTPAdapter(max_retries=retries))

    return session


def _resource_to_filename(resource: str) -> str:
    """
    Converts a `resource` into a hashed filename in a repeatable way.

    Code reference [AllenNLP](https://github.com/allenai/allennlp/blob/e5d332a592a8624e1f4ee7a9a7d30a90991db83c/allennlp/common/file_utils.py#L121)
    
    # Parameters

    resource: `str`
        The filename to be hashed.

    # Returns

    `str`
    """
    resource_bytes = resource.encode("utf-8")
    resource_hash = sha256(resource_bytes)
    filename = resource_hash.hexdigest()

    return filename


def download_url_file(url: str) -> str:
    '''
    Returns a path to the contents download from the `url`.

    This function will first check if the downloaded content already exists
    based on a cached file within the :var:`pymusas.config.PYMUSAS_CACHE_HOME` directory.
    If it does then the cached file path will be returned else the the content
    will be downloaded and cached.

    Code reference [AllenNLP](https://github.com/allenai/allennlp/blob/e5d332a592a8624e1f4ee7a9a7d30a90991db83c/allennlp/common/file_utils.py#L536)

    # Parameters

    url: `str`
        The URL address to the file to be downloaded.
 
    # Returns

    `str`
    '''
    cache_dir = config.PYMUSAS_CACHE_HOME
    os.makedirs(cache_dir, exist_ok=True)
    filename = _resource_to_filename(url)
    download_file_path = Path(cache_dir, filename)
    if download_file_path.exists():
        return str(download_file_path)
    
    with _session_with_backoff() as session:
        req = session.get(url, stream=True, timeout=5)
        req.raise_for_status()
        content_length = req.headers.get("Content-Length")
        total = int(content_length) if content_length is not None else None
        with download_file_path.open('wb') as download_file:
            with tqdm(unit="B", unit_scale=True, total=total, desc="downloading") as progress:
                for chunk in req.iter_content(chunk_size=1024):
                    if chunk:  # filter out keep-alive new chunks
                        progress.update(len(chunk))
                        download_file.write(chunk)
    return str(download_file_path)



'''
spaCy registered functions for creating the following tagger rules:
* :class:`pymusas.taggers.rules.single_word.SingleWordRule`
* :class:`pymusas.taggers.rules.mwe.MWERule`

And helper functions for the rules.
'''
from typing import Dict, List, Optional

import spacy

from pymusas.taggers.rules.mwe import MWERule
from pymusas.taggers.rules.rule import Rule
from pymusas.taggers.rules.single_word import SingleWordRule


@spacy.util.registry.misc('pymusas.taggers.rules.SingleWordRule.v1')
def single_word_rule(lexicon_collection: Dict[str, List[str]],
                     lemma_lexicon_collection: Dict[str, List[str]],
                     pos_mapper: Optional[Dict[str, List[str]]] = None
                     ) -> SingleWordRule:
    '''
    `pymusas.taggers.rules.SingleWordRule.v1` is a registered function under the
    `@misc` function register.

    See the :class:`pymusas.taggers.rules.single_word.SingleWordRule` for
    details on parameters to this function.

    # Returns
    
    :class:`pymusas.taggers.rules.single_word.SingleWordRule`
    '''
    return SingleWordRule(lexicon_collection, lemma_lexicon_collection,
                          pos_mapper)


@spacy.util.registry.misc('pymusas.taggers.rules.MWERule.v1')
def mwe_rule(mwe_lexicon_lookup: Dict[str, List[str]],
             pos_mapper: Optional[Dict[str, List[str]]] = None
             ) -> MWERule:
    '''
    `pymusas.taggers.rules.MWERule.v1` is a registered function under the
    `@misc` function register.

    See the :class:`pymusas.taggers.rules.mwe.MWERule` for details on
    parameters to this function.

    # Returns

    :class:`pymusas.taggers.rules.mwe.MWERule`
    '''
    return MWERule(mwe_lexicon_lookup, pos_mapper)


@spacy.util.registry.misc('pymusas.taggers.rules.rule_list')
def rule_list(*rules: Rule) -> List[Rule]:
    '''
    `pymusas.taggers.rules.rule_list` is a registered function under the
    `@misc` function register. The function is required when wanting to create
    a `List` of rules within a
    [config file](https://thinc.ai/docs/usage-config). We
    found it not possible to specify a `List` of custom objects within a config
    file, but is possible when using
    [variable position arguments](https://thinc.ai/docs/usage-config#registries-args),
    which this function accepts as input.
    
    This function is most likely to be
    used when creating a :class:`pymusas.spacy_api.taggers.rule_based.RuleBasedTagger`.


    # Parameters
    
    rules : `Rule`
        The :class:`pymusas.taggers.rules.rule.Rule`s to convert into a `List`
        of `Rule`s.
    
    # Returns
    
    `List[Rule]`
    '''
    return list(rules)



'''
spaCy registered functions for creating the following rankers:
* :class:`pymusas.rankers.lexicon_entry.ContextualRuleBasedRanker`
'''
from typing import List

import spacy

from pymusas.rankers.lexicon_entry import ContextualRuleBasedRanker
from pymusas.taggers.rules.rule import Rule


@spacy.util.registry.misc('pymusas.rankers.ContextualRuleBasedRanker.v1')
def contextual_rule_based_ranker(rules: List[Rule]) -> ContextualRuleBasedRanker:
    '''
    `pymusas.rankers.ContextualRuleBasedRanker.v1` is a registered function
    under the `@misc` function register.

    The parameters of this function are passed to the
    :func:`pymusas.rankers.lexicon_entry.get_construction_arguments`
    function of which the output of this function is then used as arguments
    to the :class:`pymusas.rankers.lexicon_entry.ContextualRuleBasedRanker`
    constructor.

    # Parameters

    rules : `List[Rule]`
        A `List` of :class:`pymusas.taggers.rules.rule.Rule`.

    # Returns
    
    :class:`pymusas.rankers.lexicon_entry.ContextualRuleBasedRanker`
    '''
    maximum_n_gram_length, maximum_number_wildcards \
        = ContextualRuleBasedRanker.get_construction_arguments(rules)
    return ContextualRuleBasedRanker(maximum_n_gram_length,
                                     maximum_number_wildcards)



from abc import abstractmethod
import collections
from typing import DefaultDict, Dict, List, Optional, Set, Tuple, cast

import srsly

from pymusas.base import Serialise
from pymusas.lexicon_collection import LexiconType
from pymusas.rankers.ranking_meta_data import RankingMetaData
from pymusas.taggers.rules.mwe import MWERule
from pymusas.taggers.rules.rule import Rule
from pymusas.taggers.rules.single_word import SingleWordRule


class LexiconEntryRanker(Serialise):
    '''
    An **abstract class** that defines the basic methods, `__call__`,
    `to_bytes`, and `from_bytes`, that is required for all
    :class:`LexiconEntryRanker`s.

    Each lexicon entry match is represented by a
    :class:`pymusas.rankers.ranking_meta_data.RankingMetaData` object.

    **Lower ranked lexicon entry matches should be given priority when making
    tagging decisions. A rank of 0 is better than a rank of 1.**

    A LexcionEntryRanker when called, `__call__`, returns a tuple of two `List`s
    whereby each entry in the list corresponds to a token:

    1. Contains the ranks of the lexicon entry matches as a `List[int]`.
    **Note** that the `List` can be empty if a token has no lexicon entry matches.
    2. An `Optional[RankingMetaData]` that is the global lowest ranked entry
    match for that token. If the value is `None` then no global lowest ranked
    entry can be found for that token. If the `RankingMetaData` represents more
    than one token, like a Multi Word Expression (MWE) match, then those associated tokens
    will have the same `RankingMetaData` object as the global lowest ranked entry match.
    
    **The tagger will have to make a decision how to handle global lowest ranked
    matches of value `None`, a suggested approach would be to assign an
    unmatched/unknown semantic tag to those tokens.**

    The reason for the adding the second list is that the **global** lowest
    ranked match is not the same as the local/token lowest ranked match, this is
    due to the potential of overlapping matches, e.g. `North East London brewery`
    can have a match of `North East`, `North`, and `East London brewery` in this
    case the lowest rank for `North` would be `North East`, but as we have a
    lower match that uses `East` which is `East London brewery` then the
    **global** lowest rank for `North` would be `North`.
    '''
    
    @abstractmethod
    def __call__(self, token_ranking_data: List[List[RankingMetaData]]
                 ) -> Tuple[List[List[int]], List[Optional[RankingMetaData]]]:
        '''
        For each token it returns a `List` of rankings for each lexicon entry
        match and the optional :class:`pymusas.rankers.ranking_meta_data.RankingMetaData`
        object of the **global** lowest ranked match for each token.

        # Parameters

        token_ranking_data : `List[List[RankingMetaData]]`
            For each token a `List` of :class:`pymusas.rankers.ranking_meta_data.RankingMetaData`
            representing the lexicon entry match.

        # Returns
        
        `Tuple[List[List[int]], List[Optional[RankingMetaData]]]`
        '''
        ...  # pragma: no cover

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        ...  # pragma: no cover


class ContextualRuleBasedRanker(LexiconEntryRanker):
    '''
    The contextual rule based ranker creates ranks based on the rules stated below.
    
    Each lexicon entry match is represented by a
    :class:`pymusas.rankers.ranking_meta_data.RankingMetaData` object.

    **Lower ranked lexicon entry matches should be given priority when making
    tagging decisions. See the :class:`LexiconEntryRanker` class docstring for
    more details on the returned value of the `__call__` method.**

    **Ranking Rules:**

    The ranking of lexicon entires is based off the following rules, these rules
    are based on the 6 heuristic stated at the top of column 2 on page 4 of
    [Piao et al. 2003](https://aclanthology.org/W03-1807.pdf):

    First we create an initial ranking based on lexicon entry type:

    1. Multi Word Expression (MWE) entries ranked lower than single and Non-Special
    entries are ranked lower than wild card entires.

    Then within these rankings we further rank based on:
    
    2. Longer entries, based on n-gram length, are ranked lower.
    3. Entries with fewer wildcards are ranked lower.

    Then we apply the following contextual ranking rules:

    4. Whether the POS information was excluded in the match if so these are ranked
    higher. This is only `True` when the match ignores the POS information for
    single word lexicon entries. This is always `False` when used in a
    MWE lexicon entry match.
    5. Whether the lexicon entry was matched on Token < Lemma <
    Lower cased token < Lower cased lemma. Token is the lowest ranked and lower
    cased lemma is highest.
    6. The lexicon entry that first appears in the text is ranked lowest,
    this is required for matches that do not apply to the same sequence
    of tokens.

    In the case whereby the global lowest ranked lexicon entry match is joint
    ranked with another entry then it is random which lexicon entry match is chosen.

    # Parameters

    maximum_n_gram_length : `int`
        The largest n_gram rule match that will be encountered, e.g. a match
        of `ski_noun boot_noun` will have a n-gram length of 2.
    maximum_number_wildcards : `int`
        The number of wildcards in the rule that contains the most wildcards, e.g.
        the rule `ski_* *_noun` would contain 2 wildcards. This can be 0 if you
        have no wildcard rules.

    # Instance Attributes

    n_gram_number_indexes : `int`
        The number of indexes that each n-gram length value should have when
        converting the n-gram length to a string using
        `pymusas.rankers.lexicon_entry.ContextualRuleBasedRanker.int_2_str`.
    wildcards_number_indexes : `int`
        The number of indexes that each wildcard count value should have when
        converting the wildcard count value to a string using
        `pymusas.rankers.lexicon_entry.ContextualRuleBasedRanker.int_2_str`.
    n_gram_ranking_dictionary : `Dict[int, int]`
        Maps the n-gram length to it's rank value, as the n-gram length is
        inverse to it's rank, as the larger the n-gram length the lower it's
        rank.
    '''

    def __init__(self, maximum_n_gram_length: int,
                 maximum_number_wildcards: int) -> None:

        self._maximum_n_gram_length = maximum_n_gram_length
        self._maximum_number_wildcards = maximum_number_wildcards
        
        self.n_gram_number_indexes = len(str(maximum_n_gram_length))
        self.wildcards_number_indexes = len(str(maximum_number_wildcards))

        self.n_gram_ranking_dictionary: Dict[int, int] = \
            dict(zip(range(1, maximum_n_gram_length + 1, 1),
                     range(maximum_n_gram_length, 0, -1)))

    def to_bytes(self) -> bytes:
        '''
        Serialises the :class:`ContextualRuleBasedRanker` to a bytestring.

        # Returns

        `bytes`
        '''
        serialise = {}
        serialise['maximum_n_gram_length'] = srsly.msgpack_dumps(self._maximum_n_gram_length)
        serialise['maximum_number_wildcards'] = srsly.msgpack_dumps(self._maximum_number_wildcards)
        return cast(bytes, srsly.msgpack_dumps(serialise))

    @staticmethod
    def from_bytes(bytes_data: bytes) -> "ContextualRuleBasedRanker":
        '''
        Loads :class:`ContextualRuleBasedRanker` from the given bytestring and
        returns it.

        # Parameters

        bytes_data : `bytes`
            The bytestring to load.
        
        # Returns

        :class:`ContextualRuleBasedRanker`
        '''
        serialise_data = srsly.msgpack_loads(bytes_data)
        maximum_n_gram_length = srsly.msgpack_loads(serialise_data['maximum_n_gram_length'])
        maximum_number_wildcards = srsly.msgpack_loads(serialise_data['maximum_number_wildcards'])
        return ContextualRuleBasedRanker(maximum_n_gram_length,
                                         maximum_number_wildcards)

    @staticmethod
    def get_construction_arguments(rules: List['Rule']) -> Tuple[int, int]:
        '''
        Given a `List` of rules it will return the `maximum_n_gram_length` and
        `maximum_number_wildcards` from the lexicon collections that those
        :class:`pymusas.taggers.rules.rule.Rule`(s) are based on. The output from
        this function can then be used as the arguments to the constructor of
        :class:`ContextualRuleBasedRanker`.

        # Parameters

        rules : `List[Rule]`
            A `List` of rules. This `List` is typically required when creating
            a :class:`pymusas.taggers.rule_based.RuleBasedTagger` tagger.
        
        # Returns

        `Tuple[int, int]`

        # Examples
        ``` python
        >>> from pymusas.rankers.lexicon_entry import ContextualRuleBasedRanker
        >>> from pymusas.taggers.rules.mwe import MWERule
        >>> from pymusas.lexicon_collection import MWELexiconCollection
        >>> pt_mwe_lexicon_url = "https://raw.githubusercontent.com/UCREL/Multilingual-USAS/master/Portuguese/mwe-pt.tsv"
        >>> mwe_dict = MWELexiconCollection.from_tsv(pt_mwe_lexicon_url)
        >>> mwe_rule = MWERule(mwe_dict)
        >>> ranker_construction_arguments = ContextualRuleBasedRanker.get_construction_arguments([mwe_rule])
        >>> ranker = ContextualRuleBasedRanker(*ranker_construction_arguments)

        ```
        '''
        maximum_n_gram_length = 0
        maximum_number_wildcards = 0
        for rule in rules:
            rule_max_n_gram_length = 0
            rule_max_number_wildcards = 0
            if isinstance(rule, SingleWordRule):
                rule_max_n_gram_length = 1
            elif isinstance(rule, MWERule):
                rule_max_n_gram_length = rule.mwe_lexicon_collection.longest_mwe_template
                rule_max_number_wildcards = rule.mwe_lexicon_collection.most_wildcards_in_mwe_template

            if rule_max_n_gram_length > maximum_n_gram_length:
                maximum_n_gram_length = rule_max_n_gram_length

            if rule_max_number_wildcards > maximum_number_wildcards:
                maximum_number_wildcards = rule_max_number_wildcards

        return (maximum_n_gram_length, maximum_number_wildcards)
        
    @staticmethod
    def int_2_str(int_value: int, number_indexes: int) -> str:
        '''
        Converts the integer, `int_value`, to a string with `number_indexes`,
        e.g. `10` and `05` both have `number_indexes` of 2 and `001`, `020`,
        and `211` have `number_indexes` of 3.

        # Parameters

        int_value : `int`
            The integer to converts to a string with the given `number_indexes`.
        number_indexes : `int`
            The number of indexes the `int_value` should have in the returned
            string.
        
        # Returns
        
        `str`

        # Raises

        ValueError
            If the `number_indexes` of the `int_value` when converted to a
            string is greater than the given `number_indexes`.
        '''
        str_value = str(int_value)
        str_value_number_indexes = len(str_value)
        if str_value_number_indexes > number_indexes:
            error_msg = (f"Cannot convert int ({int_value}) to a ranked"
                         f" string as the maximum number of indexes it can be"
                         f" is {number_indexes}.")
            raise ValueError(error_msg)
        
        number_prefix_zeros_to_add = number_indexes - str_value_number_indexes
        prefix_zeros = '0' * number_prefix_zeros_to_add
        return f'{prefix_zeros}{str_value}'

    @staticmethod
    def get_global_lowest_ranks(token_ranking_data: List[List[RankingMetaData]],
                                token_rankings: List[List[int]],
                                ranking_data_to_exclude: Optional[Set[RankingMetaData]] = None
                                ) -> List[Optional[RankingMetaData]]:
        '''
        Returns the global lowest ranked entry match for each token. If the value
        is `None` then no global lowest ranked entry can be found for that token.
        If the `RankingMetaData` represents more than one token, like a Multi
        Word Expression (MWE) match, then those associated tokens will have the
        same `RankingMetaData` object as the global lowest ranked entry match.

        Time Complexity, given *N* is the number of tokens, *M* is the number
        of unique ranking data, and *P* is the number of ranking data (non-unique)
        then the time complexity is:
        
        O(N + P) + O(M log M) + O(M)
        
        # Parameters

        token_ranking_data : `List[List[RankingMetaData]]`
            For each token a `List` of
            :class:`pymusas.rankers.ranking_meta_data.RankingMetaData` representing
            the lexicon entry match.
        token_rankings : `List[List[int]]`
            For each token contains the ranks of the lexicon entry matches.
            **Note** that the `List` can be empty if a token has no lexicon
            entry matches.
        ranking_data_to_exclude : `Set[RankingMetaData]`, optional (default = `None`)
            Any :class:`pymusas.rankers.ranking_meta_data.RankingMetaData` to
            exclude from the ranking selection, this can be useful when wanting
            to get the next best global rank for each token.

        # Raises

        `AssertionError`
            If the length of `token_ranking_data` is not equal to the length of
            `token_rankings`, for both the outer and inner `List`s.

        # Examples
        ``` python
        >>> from pymusas.rankers.lexicon_entry import ContextualRuleBasedRanker
        >>> from pymusas.rankers.ranking_meta_data import RankingMetaData
        >>> from pymusas.lexicon_collection import LexiconType
        >>> from pymusas.rankers.lexical_match import LexicalMatch
        >>> north_east = RankingMetaData(LexiconType.MWE_NON_SPECIAL, 2, 0,
        ...                              False, LexicalMatch.TOKEN, 0, 2,
        ...                              'North_noun East_noun', ('Z1',))
        >>> east_london_brewery = RankingMetaData(LexiconType.MWE_NON_SPECIAL, 3, 0,
        ...                                       False, LexicalMatch.TOKEN, 1, 4,
        ...                                       'East_noun London_noun brewery_noun', ('Z1',))
        >>> token_ranking_data = [
        ...     [
        ...         north_east
        ...     ],
        ...     [
        ...         north_east,
        ...         east_london_brewery
        ...     ],
        ...     [
        ...         east_london_brewery
        ...     ],
        ...     [
        ...         east_london_brewery
        ...     ]
        ... ]
        >>> token_rankings = [[120110], [120110, 110111], [110111], [110111]]
        >>> expected_lowest_ranked_matches = [None, east_london_brewery,
        ...                                   east_london_brewery, east_london_brewery]
        >>> assert (ContextualRuleBasedRanker.get_global_lowest_ranks(token_ranking_data, token_rankings, None)
        ...         == expected_lowest_ranked_matches)

        ```

        Following on from the previous example, we now want to find the next best
        global match for each token so we exclude the current best global match
        for each token which is the `east_london_brewery` match:
        
        ``` python
        >>> expected_lowest_ranked_matches = [north_east, north_east, None, None]
        >>> ranking_data_to_exclude = {east_london_brewery}
        >>> assert (ContextualRuleBasedRanker.get_global_lowest_ranks(token_ranking_data, token_rankings,
        ...                                                          ranking_data_to_exclude)
        ...         == expected_lowest_ranked_matches)

        ```
        '''
        if ranking_data_to_exclude is None:
            ranking_data_to_exclude = set()
        
        assert len(token_ranking_data) == len(token_rankings), 'Lengths should be equal'

        ranking_meta_data: DefaultDict[int, Set[RankingMetaData]] = collections.defaultdict(set)
        for token_data, token_ranking in zip(token_ranking_data, token_rankings):
            assert len(token_data) == len(token_ranking), 'Lengths should be equal'
            
            for data, rank in zip(token_data, token_ranking):
                if data in ranking_data_to_exclude:
                    continue
                ranking_meta_data[rank].add(data)
        
        global_lowest_ranks: List[Optional[RankingMetaData]] = \
            [None for _ in token_ranking_data]
        
        ordered_ranking_meta_data = sorted(ranking_meta_data.items(),
                                           key=lambda x: x[0])
        for rank, meta_data in ordered_ranking_meta_data:
            for data in meta_data:
                start, end = data.token_match_start_index, data.token_match_end_index
                if any(global_lowest_ranks[start: end]):
                    continue

                for index in range(start, end):
                    global_lowest_ranks[index] = data

        return global_lowest_ranks

    def __call__(self, token_ranking_data: List[List[RankingMetaData]]
                 ) -> Tuple[List[List[int]], List[Optional[RankingMetaData]]]:
        '''
        For each token it returns a `List` of rankings for each lexicon entry
        match and the optional :class:`pymusas.rankers.ranking_meta_data.RankingMetaData`
        object of the **global** lowest ranked match for each token.
        
        See the ranking rules in the class docstring for details on how
        each lexicon entry match is ranked.

        Time Complexity, given *N* is the number of tokens, *M* is the number
        of unique ranking data, and *P* is the number of ranking data (non-unique)
        then the time complexity is:
        
        O(3(N + P)) + O(M log M) + O(M)

        # Parameters

        token_ranking_data : `List[List[RankingMetaData]]`
            For each token a `List` of :class:`pymusas.rankers.ranking_meta_data.RankingMetaData`
            representing the lexicon entry match.

        # Returns
        
        `Tuple[List[List[int]], List[Optional[RankingMetaData]]]`

        # Examples
        ```python
        >>> from pymusas.rankers.lexicon_entry import ContextualRuleBasedRanker
        >>> from pymusas.rankers.ranking_meta_data import RankingMetaData
        >>> from pymusas.lexicon_collection import LexiconType
        >>> from pymusas.rankers.lexical_match import LexicalMatch
        >>> north_east = RankingMetaData(LexiconType.MWE_NON_SPECIAL, 2, 0,
        ...                              False, LexicalMatch.TOKEN, 0, 2,
        ...                              'North_noun East_noun', ('Z1',))
        >>> east_london_brewery = RankingMetaData(LexiconType.MWE_NON_SPECIAL, 3, 0,
        ...                                       False, LexicalMatch.TOKEN, 1, 4,
        ...                                       'East_noun London_noun brewery_noun', ('Z1',))
        >>> token_ranking_data = [
        ...     [
        ...         north_east
        ...     ],
        ...     [
        ...         north_east,
        ...         east_london_brewery
        ...     ],
        ...     [
        ...         east_london_brewery
        ...     ],
        ...     [
        ...         east_london_brewery
        ...     ]
        ... ]
        >>> expected_ranks = [[120110], [120110, 110111], [110111], [110111]]
        >>> expected_lowest_ranked_matches = [None, east_london_brewery,
        ...                                   east_london_brewery, east_london_brewery]
        >>> ranker = ContextualRuleBasedRanker(3, 0)
        >>> assert ((expected_ranks, expected_lowest_ranked_matches)
        ...         == ranker(token_ranking_data))

        ```
        '''
        lexicon_type_to_rank = {
            LexiconType.MWE_NON_SPECIAL: 1,
            LexiconType.MWE_WILDCARD: 2,
            LexiconType.MWE_CURLY_BRACES: 3,
            LexiconType.SINGLE_NON_SPECIAL: 4
        }

        exclude_pos_information_to_rank = {
            False: 1,
            True: 2
        }

        initial_rankings: List[List[str]] = []
        largest_token_index = 0
        for token in token_ranking_data:
            token_rankings: List[str] = []
            for ranking_data in token:
                lexicon_type_rank = lexicon_type_to_rank[ranking_data.lexicon_type]
                n_gram_rank = self.n_gram_ranking_dictionary[ranking_data.lexicon_n_gram_length]
                n_gram_str_rank = self.int_2_str(n_gram_rank, self.n_gram_number_indexes)
                wildcard_str_rank = self.int_2_str(ranking_data.lexicon_wildcard_count,
                                                   self.wildcards_number_indexes)
                exclude_pos_information_rank = exclude_pos_information_to_rank[ranking_data.exclude_pos_information]
                lexical_match_rank = ranking_data.lexical_match.value
                rank_str = (f'{lexicon_type_rank}{n_gram_str_rank}{wildcard_str_rank}'
                            f'{exclude_pos_information_rank}{lexical_match_rank}')
                token_rankings.append(rank_str)

                if largest_token_index < ranking_data.token_match_start_index:
                    largest_token_index = ranking_data.token_match_start_index
                if largest_token_index < ranking_data.token_match_end_index:
                    largest_token_index = ranking_data.token_match_end_index
            initial_rankings.append(token_rankings)

        # Add to each token ranking where it first appears in the text, rule 6.
        largest_token_index_number_indexes = len(str(largest_token_index))
        rankings: List[List[int]] = []
        for str_token_rankings, token in zip(initial_rankings, token_ranking_data):
            int_token_rankings: List[int] = []
            for str_ranking, ranking_data in zip(str_token_rankings, token):
                start_index_str_rank = self.int_2_str(ranking_data.token_match_start_index,
                                                      largest_token_index_number_indexes)
                int_token_ranking = int(f'{str_ranking}{start_index_str_rank}')
                int_token_rankings.append(int_token_ranking)
            rankings.append(int_token_rankings)
        global_lowest_rank_indexes = self.get_global_lowest_ranks(token_ranking_data,
                                                                  rankings, None)

        return (rankings, global_lowest_rank_indexes)

    def __eq__(self, other: object) -> bool:
        '''
        Given another object to compare too it will return `True` if the other
        object is the same class and was initialised using with the same
        `maximum_n_gram_length` and `maximum_number_wildcards` values.

        # Parameters

        other : `object`
            The object to compare too.
        
        # Returns

        `True`
        '''
        if not isinstance(other, ContextualRuleBasedRanker):
            return False
        
        if self._maximum_n_gram_length != other._maximum_n_gram_length:
            return False

        if self._maximum_number_wildcards != other._maximum_number_wildcards:
            return False

        return True



from typing import Dict, List, Optional, Tuple

from pymusas.lexicon_collection import LexiconType, MWELexiconCollection
from pymusas.rankers.lexical_match import LexicalMatch
from pymusas.rankers.ranking_meta_data import RankingMetaData
from pymusas.taggers.rules import util
from pymusas.taggers.rules.rule import Rule


class MWERule(Rule):
    '''
    A Multi Word Expression (MWE) rule match can be one of the following matches:

    1. `MWE_NON_SPECIAL` match - whereby the combined token/lemma and POS
    is found within the given MWE Lexicon Collection (`self.mwe_lexicon_collection`).
    2. `MWE_WILDCARD` match - whereby the combined token/lemma and POS matches
    a wildcard MWE template that is within the MWE Lexicon Collection
    (`self.mwe_lexicon_collection`).

    All rule matches use the
    `pymusas.lexicon_collection.MWELexiconCollection.mwe_match`
    method for matching. Matches are found based on the original token/lemma and
    lower cased versions of the token/lemma.

    # Parameters

    mwe_lexicon_lookup : `Dict[str, List[str]]`
        The data to create `mwe_lexicon_collection` instance attribute. A
        Dictionary where the keys are MWE templates, of any
        :class:`pymusas.lexicon_collection.LexiconType`,
        and the values are a list of associated semantic tags.
    pos_mapper : `Dict[str, List[str]]`, optional (default = `None`)
        If not `None`, maps from the `mwe_lexicon_lookup` POS tagset to the
        desired POS tagset,whereby the mapping is a `List` of tags,
        at the moment there is no preference order in this list of POS tags.
        **Note** the longer the `List[str]` for
        each POS mapping the slower the tagger, a one to one mapping will have
        no speed impact on the tagger. A selection of POS mappers can be found in
        :mod:`pymusas.pos_mapper`.

    # Instance Attributes

    mwe_lexicon_collection : `pymusas.lexicon_collection.MWELexiconCollection`
        A :class:`pymusas.lexicon_collection.MWELexiconCollection` instance that
        has been initialised using the `mwe_lexicon_lookup` and `pos_mapper`
        parameters. This collection is used to find MWE rule matches.
    '''

    def __init__(self, mwe_lexicon_lookup: Dict[str, List[str]],
                 pos_mapper: Optional[Dict[str, List[str]]] = None) -> None:
        
        self.mwe_lexicon_collection = MWELexiconCollection(mwe_lexicon_lookup,
                                                           pos_mapper)

    def __call__(self, tokens: List[str], lemmas: List[str], pos_tags: List[str]
                 ) -> List[List[RankingMetaData]]:
        '''
        Given the tokens, lemmas, and POS tags for each word in a text,
        it returns for each token a `List` of rules matches defined by
        the :class:`pymusas.rankers.ranking_meta_data.RankingMetaData` object based on
        the rule matches stated in the class docstring above.
        
        # Parameters

        tokens : `List[str]`
            The tokens that are within the text.
        lemmas : `List[str]`
            The lemmas of the tokens.
        pos_tags : `List[str]`
            The Part Of Speech tags of the tokens.
        
        # Returns

        `List[List[RankingMetaData]]`
        '''

        def tag_n_gram_indexes(_n_gram_indexes: List[Tuple[int, int]],
                               mwe_type: LexiconType,
                               token_pos: List[str], lemma_pos: List[str],
                               token_lower_pos: List[str], lemma_lower_pos: List[str],
                               token_ranking_meta_data: List[List[RankingMetaData]]
                               ) -> None:
            '''
            Given a `List` of n gram indexes, whereby all grams have the same value
            of *n*, it will find all MWE rule matches of type `mwe_type` for those
            n-grams in the four token `List`s (`token_pos`, `lemma_pos`,
            `token_lower_pos`, `lemma_lower_pos`). All rule matches will crate a
            `RankingMetaData` object for each token in the match, of which these
            objects are appended to the relevant token indexes within the
            `token_ranking_meta_data` object.

            Nothing is returned but the `token_ranking_meta_data` object will have
            been updated if at least one rule match has been found.
            '''
            token_list_type: List[Tuple[List[str], LexicalMatch]]
            token_list_type = [(token_pos, LexicalMatch.TOKEN),
                               (lemma_pos, LexicalMatch.LEMMA),
                               (token_lower_pos, LexicalMatch.TOKEN_LOWER),
                               (lemma_lower_pos, LexicalMatch.LEMMA_LOWER)]
            
            for token_list, token_type in token_list_type:
                for n_gram_index in _n_gram_indexes:
                    start_index, end_index = n_gram_index
                    mwe_template = ''
                    for token_index in range(start_index, end_index):
                        if (token_index + 1) == end_index:
                            mwe_template += f'{token_list[token_index]}'
                            continue
                        mwe_template += f'{token_list[token_index]} '
                    matched_mwe_templates = self.mwe_lexicon_collection.mwe_match(mwe_template,
                                                                                  mwe_type)
                    if not matched_mwe_templates:
                        continue

                    for matched_mwe_template in matched_mwe_templates:
                        mwe_meta_data = self.mwe_lexicon_collection[matched_mwe_template]
                        semantic_tags = tuple(mwe_meta_data.semantic_tags)
                        ranking_meta_data = RankingMetaData(mwe_type,
                                                            mwe_meta_data.n_gram_length,
                                                            mwe_meta_data.wildcard_count,
                                                            False, token_type, start_index,
                                                            end_index,
                                                            matched_mwe_template,
                                                            semantic_tags)
                        for token_index in range(start_index, end_index):
                            token_ranking_meta_data[token_index].append(ranking_meta_data)
                
        def tag_n_gram_based_MWE(mwe_type: LexiconType,
                                 token_pos: List[str], lemma_pos: List[str],
                                 token_lower_pos: List[str], lemma_lower_pos: List[str],
                                 token_ranking_meta_data: List[List[RankingMetaData]]
                                 ) -> None:
            '''
            Any type of rule match that uses n-grams should use this function.
            At the moment only the `MWE_NON_SPECIAL` and `MWE_WILDCARD` matches
            are supported.
            
            The function for each token list () creates n-grams, up to the size
            of the maximum n-gram for the match type requested, of these tokens
            and then matches those n-grams using the
            `pymusas.lexicon_collection.MWELexiconCollection.mwe_match`
            method.

            Nothing is returned but the `token_ranking_meta_data` object will have
            been updated if at least one rule match has been found.
            '''
            largest_mwe_in_lexicon = 0
            if mwe_type == LexiconType.MWE_NON_SPECIAL:
                largest_mwe_in_lexicon = self.mwe_lexicon_collection.longest_non_special_mwe_template
            elif mwe_type == LexiconType.MWE_WILDCARD:
                largest_mwe_in_lexicon = self.mwe_lexicon_collection.longest_wildcard_mwe_template
            
            # If we do not have any MWE lexicons that are at least 2 grams
            # then return
            if largest_mwe_in_lexicon < 2:
                return
            
            largest_n_gram = largest_mwe_in_lexicon + 1
            largest_n_gram_indexes: List[Tuple[int, int]] = []

            for n_gram_index in util.n_gram_indexes(token_pos, 2, largest_mwe_in_lexicon):
                start_index, end_index = n_gram_index
                n_gram_size = end_index - start_index
                if largest_n_gram > n_gram_size:
                    largest_n_gram = n_gram_size
                    tag_n_gram_indexes(largest_n_gram_indexes, mwe_type,
                                       token_pos, lemma_pos, token_lower_pos,
                                       lemma_lower_pos,
                                       token_ranking_meta_data)
                    largest_n_gram_indexes = []

                largest_n_gram_indexes.append(n_gram_index)
            
            tag_n_gram_indexes(largest_n_gram_indexes, mwe_type, token_pos,
                               lemma_pos, token_lower_pos,
                               lemma_lower_pos, token_ranking_meta_data)
        
        token_pos: List[str] = []
        token_lower_pos: List[str] = []
        lemma_pos: List[str] = []
        lemma_lower_pos: List[str] = []
        for token, lemma, pos in zip(tokens, lemmas, pos_tags):
            token_pos.append(f'{token}_{pos}')
            token_lower_pos.append(f'{token}_{pos}'.lower())
            lemma_pos.append(f'{lemma}_{pos}')
            lemma_lower_pos.append(f'{lemma}_{pos}'.lower())
        
        number_tokens = len(tokens)
        token_ranking_meta_data: List[List[RankingMetaData]] \
            = [[] for _ in range(number_tokens)]
        
        if number_tokens < 2:
            return token_ranking_meta_data

        # First match on `MWE_NON_SPECIAL`
        tag_n_gram_based_MWE(LexiconType.MWE_NON_SPECIAL, token_pos, lemma_pos,
                             token_lower_pos, lemma_lower_pos,
                             token_ranking_meta_data)

        # Second match on `MWE_WILDCARD`
        tag_n_gram_based_MWE(LexiconType.MWE_WILDCARD, token_pos, lemma_pos,
                             token_lower_pos, lemma_lower_pos,
                             token_ranking_meta_data)

        return token_ranking_meta_data

    def to_bytes(self) -> bytes:
        '''
        Serialises the :class:`MWERule` to a bytestring.

        # Returns

        `bytes`
        '''
        return self.mwe_lexicon_collection.to_bytes()

    @staticmethod
    def from_bytes(bytes_data: bytes) -> "MWERule":
        '''
        Loads :class:`MWERule` from the given bytestring and returns it.

        # Parameters

        bytes_data : `bytes`
            The bytestring to load.
        
        # Returns

        :class:`MWERule`
        '''
        mwe_rule = MWERule({}, None)
        mwe_lexicon_collection = MWELexiconCollection.from_bytes(bytes_data)
        mwe_rule.mwe_lexicon_collection = mwe_lexicon_collection
        return mwe_rule

    def __eq__(self, other: object) -> bool:
        '''
        Given another object to compare too it will return `True` if the other
        object is the same class and initialised using with the same argument
        values.

        # Parameters

        other : `object`
            The object to compare too.
        
        # Returns

        `True`
        '''
        if not isinstance(other, MWERule):
            return False
        
        if self.mwe_lexicon_collection != other.mwe_lexicon_collection:
            return False

        return True



'''
spaCy registered functions for loading Part Of Speech (POS) mappings.
'''

from typing import Dict, List

import spacy

from pymusas import pos_mapper


@spacy.util.registry.misc('pymusas.pos_mapper.UPOS_TO_USAS_COREv1')
def upos_to_usas_core() -> Dict[str, List[str]]:
    '''
    `pymusas.pos_mapper.UPOS_TO_USAS_COREv1` is a registered function under the
    `@misc` function register. It returns the
    :var:`pymusas.pos_mapper.UPOS_TO_USAS_CORE` mapping.

    # Returns

    `Dict[str, List[str]]`
    '''
    return pos_mapper.UPOS_TO_USAS_CORE


@spacy.util.registry.misc('pymusas.pos_mapper.USAS_CORE_TO_UPOSv1')
def usas_core_to_upos() -> Dict[str, List[str]]:
    '''
    `pymusas.pos_mapper.USAS_CORE_TO_UPOSv1` is a registered function under the
    `@misc` function register. It returns the
    :var:`pymusas.pos_mapper.USAS_CORE_TO_UPOS` mapping.

    # Returns

    `Dict[str, List[str]]`
    '''
    return pos_mapper.USAS_CORE_TO_UPOS


@spacy.util.registry.misc('pymusas.pos_mapper.PENN_CHINESE_TREEBANK_TO_USAS_COREv1')
def penn_chinese_treebank_to_usas_core() -> Dict[str, List[str]]:
    '''
    `pymusas.pos_mapper.PENN_CHINESE_TREEBANK_TO_USAS_COREv1` is a registered
    function under the `@misc` function register. It returns the
    :var:`pymusas.pos_mapper.PENN_CHINESE_TREEBANK_TO_USAS_CORE` mapping.

    # Returns

    `Dict[str, List[str]]`
    '''
    return pos_mapper.PENN_CHINESE_TREEBANK_TO_USAS_CORE


@spacy.util.registry.misc('pymusas.pos_mapper.USAS_CORE_TO_PENN_CHINESE_TREEBANKv1')
def usas_core_to_penn_chinese_treebank() -> Dict[str, List[str]]:
    '''
    `pymusas.pos_mapper.USAS_CORE_TO_PENN_CHINESE_TREEBANKv1` is a registered
    function under the `@misc` function register. It returns the
    :var:`pymusas.pos_mapper.USAS_CORE_TO_PENN_CHINESE_TREEBANK` mapping.

    # Returns

    `Dict[str, List[str]]`
    '''
    return pos_mapper.USAS_CORE_TO_PENN_CHINESE_TREEBANK


@spacy.util.registry.misc('pymusas.pos_mapper.BASIC_CORCENCC_TO_USAS_COREv1')
def basic_corcencc_to_usas_core() -> Dict[str, List[str]]:
    '''
    `pymusas.pos_mapper.BASIC_CORCENCC_TO_USAS_COREv1` is a registered
    function under the `@misc` function register. It returns the
    :var:`pymusas.pos_mapper.BASIC_CORCENCC_TO_USAS_CORE` mapping.

    # Returns

    `Dict[str, List[str]]`
    '''
    return pos_mapper.BASIC_CORCENCC_TO_USAS_CORE


@spacy.util.registry.misc('pymusas.pos_mapper.USAS_CORE_TO_BASIC_CORCENCCv1')
def usas_core_to_basic_corcencc() -> Dict[str, List[str]]:
    '''
    `pymusas.pos_mapper.USAS_CORE_TO_BASIC_CORCENCCv1` is a registered
    function under the `@misc` function register. It returns the
    :var:`pymusas.pos_mapper.USAS_CORE_TO_BASIC_CORCENCC` mapping.

    # Returns

    `Dict[str, List[str]]`
    '''
    return pos_mapper.USAS_CORE_TO_BASIC_CORCENCC



import collections
from collections.abc import MutableMapping
import csv
from dataclasses import dataclass
from enum import Enum, unique
from os import PathLike
import re
import typing
from typing import DefaultDict, Dict, Generator, List, Optional, Set, Tuple, Union, cast
from urllib.parse import urlparse
import warnings

import srsly

from . import file_utils, utils


@unique
class LexiconType(str, Enum):
    '''
    Descriptions of the type associated to single and Multi Word Expression (MWE)
    lexicon entires and templates. Any type with the word `NON_SPECIAL` means
    that it does not use any special syntax, for example does not use wildcards
    or curly braces.

    The `value` attribute of each instance attribute is of type `str` describing
    the type associated with that attribute. For the best explanation see the
    example below.
    
    # Instance Attributes

    SINGLE_NON_SPECIAL : `LexiconType`
        Single word lexicon lookup.
    MWE_NON_SPECIAL : `LexiconType`
        MWE lexicon lookup.
    MWE_WILDCARD : `LexiconType`
        MWE lexicon lookup using a wildcard.
    MWE_CURLY_BRACES : `LexiconType`
        MWE lexicon lookup using curly braces.

    # Examples
    ```python
    >>> from pymusas.lexicon_collection import LexiconType
    >>> assert 'Single Non Special' == LexiconType.SINGLE_NON_SPECIAL
    >>> assert 'Single Non Special' == LexiconType.SINGLE_NON_SPECIAL.value
    >>> assert 'SINGLE_NON_SPECIAL' == LexiconType.SINGLE_NON_SPECIAL.name
    >>> all_possible_values = {'Single Non Special', 'MWE Non Special',
    ... 'MWE Wildcard', 'MWE Curly Braces'}
    >>> assert all_possible_values == {lexicon_type.value for lexicon_type in LexiconType}
    
    ```
    '''
    SINGLE_NON_SPECIAL = 'Single Non Special'
    MWE_NON_SPECIAL = 'MWE Non Special'
    MWE_WILDCARD = 'MWE Wildcard'
    MWE_CURLY_BRACES = 'MWE Curly Braces'

    def __repr__(self) -> str:
        '''
        Machine readable string. When printed and run `eval()` over the string
        you should be able to recreate the object.
        '''
        return self.__str__()


@dataclass(init=True, repr=True, eq=True, order=False,
           unsafe_hash=False, frozen=True)
class LexiconEntry:
    '''
    A LexiconEntry contains the `semantic_tags` that are associated with a
    `lemma` and optionally the lemma's `POS`.

    As frozen is true, the attributes cannot be assigned another value.

    This data type is mainly used for single word lexicons, rather than
    Multi Word Expression (MWE).

    **Note** the parameters to the `__init__` are the same as the Instance
    Attributes.

    # Instance Attributes

    lemma: `str`
        The lemma of a token or the token itself.
    semantic_tags: `List[str]`
        The semantic tags associated with the `lemma` and optional `POS`.
        The semantic tags are in rank order, the most likely tag
        is the first tag in the list.
    pos: `str`, optional (default = `None`)
        The Part Of Speech (POS) to be associated with the `lemma`.
    '''

    lemma: str
    semantic_tags: List[str]
    pos: Optional[str] = None


@dataclass(init=True, repr=True, eq=True, order=False,
           unsafe_hash=False, frozen=True)
class LexiconMetaData:
    '''
    A LexiconMetaData object contains all of the meta data about a given
    single word or Multi Word Expression (MWE) lexicon entry. This meta data can
    be used to help rank single and MWE entries when tagging.

    As frozen is true, the attributes cannot be assigned another value.

    **Note** the parameters to the `__init__` are the same as the Instance
    Attributes.

    # Instance Attributes

    semantic_tags : `List[str]`
        The semantic tags associated with the lexicon entry.
        The semantic tags are in rank order, the most likely tag
        is the first tag in the list.
    n_gram_length : `int`
        The n-gram size of the lexicon entry, e.g. `*_noun boot*_noun` will be
        of length 2 and all single word lexicon entries will be of length 1.
    lexicon_type : `LexiconType`
        Type associated to the lexicon entry.
    wildcard_count : `int`
        Number of wildcards in the lexicon entry, e.g. `*_noun boot*_noun` will
        be 2 and `ski_noun boot_noun` will be 0.
    '''

    semantic_tags: List[str]
    n_gram_length: int
    lexicon_type: LexiconType
    wildcard_count: int


class LexiconCollection(MutableMapping):
    '''
    This is a dictionary object that will hold :class:`LexiconEntry` data in a fast to
    access object. The keys of the dictionary are expected to be either just a
    lemma or a combination of lemma and pos in the following format:
    `{lemma}|{pos}` e.g. `Car|Noun`.

    The value to each key is the associated semantic tags, whereby the semantic
    tags are in rank order, the most likely tag is the first tag in the list.
    
    **Note** that the `lemma` can be the token
    itself rather than just it's base form, e.g. can be `Cars` rather than `Car`.

    This data type is used for single word lexicons, to store Multi Word
    Expression (MWE) see the :class:`MWELexiconCollection`.

    # Parameters

    data: `Dict[str, List[str]]`, optional (default = `None`)

    # Instance Attributes

    data: `Dict[str, List[str]]`
        Dictionary where the keys are `{lemma}|{pos}` and the values are
        a list of associated semantic tags. If the `data` parameter given was
        `None` then the value of this attribute will be an empty dictionary.

    # Examples
    ``` python
    >>> from pymusas.lexicon_collection import LexiconEntry, LexiconCollection
    >>> lexicon_entry = LexiconEntry('London', ['Z3', 'Z1', 'A1'], 'noun')
    >>> collection = LexiconCollection()
    >>> collection.add_lexicon_entry(lexicon_entry)
    >>> most_likely_tag = collection['London|noun'][0]
    >>> assert most_likely_tag == 'Z3'
    >>> least_likely_tag = collection['London|noun'][-1]
    >>> assert least_likely_tag == 'A1'

    ```

    '''
    
    def __init__(self, data: Optional[Dict[str, List[str]]] = None) -> None:

        self.data: Dict[str, List[str]] = {}
        if data is not None:
            self.data = data

    def add_lexicon_entry(self, value: LexiconEntry,
                          include_pos: bool = True) -> None:
        '''
        Will add the :class:`LexiconEntry` to the collection, whereby the key is the
        combination of the lemma and pos and the value are the semantic tags.
        
        The lemma and pos are combined as follows: `{lemma}|{pos}`, e.g.
        `Car|Noun`

        If the pos value is None then then only the lemma is used: `{lemma}`,
        e.g. `Car`

        # Parameters

        value: `LexiconEntry`
            Lexicon Entry to add to the collection.
        include_pos: `bool`, optional (default = `True`)
            Whether to include the POS tag within the key.
        '''
        lemma = value.lemma
        if value.pos is not None and include_pos:
            lemma += f'|{value.pos}'
        self[lemma] = value.semantic_tags

    def to_dictionary(self) -> Dict[str, List[str]]:
        '''
        Returns the `data` instance attribute.

        # Returns

        `Dict[str, List[str]]`
        '''

        return self.data

    def to_bytes(self) -> bytes:
        '''
        Serialises the :class:`LexiconCollection` to a bytestring.

        # Returns

        `bytes`
        '''
        return cast(bytes, srsly.msgpack_dumps(self.data))

    @staticmethod
    def from_bytes(bytes_data: bytes) -> "LexiconCollection":
        '''
        Loads :class:`LexiconCollection` from the given bytestring and
        returns it.

        # Parameters

        bytes_data : `bytes`
            The bytestring to load.
        
        # Returns

        :class:`LexiconCollection`
        '''
        return LexiconCollection(srsly.msgpack_loads(bytes_data))

    @staticmethod
    def from_tsv(tsv_file_path: Union[PathLike, str], include_pos: bool = True
                 ) -> Dict[str, List[str]]:
        '''
        Given a `tsv_file_path` it will return a dictionary object that can
        be used to create a :class:`LexiconCollection`.

        Each line in the TSV file will be read in as a :class:`LexiconEntry`
        and added to a temporary :class:`LexiconCollection`, once all lines
        in the TSV have been parsed the return value is the `data` attribute of
        the temporary :class:`LexiconCollection`.

        If the file path is a URL, the file will be downloaded and cached using
        :func:`pymusas.file_utils.download_url_file`.
        
        If `include_pos` is True and the TSV file does not contain a
        `pos` field heading then this will return a LexiconCollection that is
        identical to a collection that ran this method with `include_pos` equal
        to False.

        Code reference, the identification of a URL and the idea to do this has
        come from the [AllenNLP library](https://github.com/allenai/allennlp/blob/main/allennlp/common/file_utils.py#L205)

        # Parameters

        tsv_file_path: `Union[PathLike, str]`
            A file path or URL to a TSV file that contains at least two
            fields, with an optional third, with the following headings:
            
            1. `lemma`,
            2. `semantic_tags`
            3. `pos` (Optional)
            
            All other fields will be ignored.
        include_pos: `bool`, optional (default = `True`)
            Whether to include the POS information, if the information is avaliable,
            or not. See :func:`add_lexicon_entry` for more information on this
            parameter.

        # Returns
        
        `Dict[str, List[str]]`
        
        # Raises
        
        `ValueError`
            If the minimum field headings, `lemma` and `semantic_tags`, do not
            exist in the given TSV file.

        # Examples

        `include_pos` = `True`
        ``` python
        >>> from pymusas.lexicon_collection import LexiconCollection
        >>> welsh_lexicon_url = 'https://raw.githubusercontent.com/apmoore1/Multilingual-USAS/master/Welsh/semantic_lexicon_cy.tsv'
        >>> welsh_lexicon_dict = LexiconCollection.from_tsv(welsh_lexicon_url, include_pos=True)
        >>> welsh_lexicon_collection = LexiconCollection(welsh_lexicon_dict)
        >>> assert welsh_lexicon_dict['ceir|noun'][0] == 'M3fn'
        >>> assert welsh_lexicon_dict['ceir|verb'][0] == 'A9+'

        ```

        `include_pos` = `False`
        ``` python
        >>> from pymusas.lexicon_collection import LexiconCollection
        >>> welsh_lexicon_url = 'https://raw.githubusercontent.com/apmoore1/Multilingual-USAS/master/Welsh/semantic_lexicon_cy.tsv'
        >>> welsh_lexicon_dict = LexiconCollection.from_tsv(welsh_lexicon_url, include_pos=False)
        >>> welsh_lexicon_collection = LexiconCollection(welsh_lexicon_dict)
        >>> assert welsh_lexicon_dict['ceir'][0] == 'M3fn'

        ```
        '''
        minimum_field_names = {'lemma', 'semantic_tags'}
        extra_field_names = ['pos']
        field_names_to_extract = []

        collection_from_tsv = LexiconCollection()

        if not isinstance(tsv_file_path, str):
            tsv_file_path = str(tsv_file_path)

        parsed = urlparse(tsv_file_path)
        if parsed.scheme in ("http", "https", "s3", "hf", "gs"):
            tsv_file_path = file_utils.download_url_file(tsv_file_path)

        with open(tsv_file_path, 'r', newline='', encoding='utf-8') as fp:
            csv_reader = csv.DictReader(fp, delimiter='\t')
            file_field_names: Set[str] = set()
            if csv_reader.fieldnames:
                file_field_names = set(csv_reader.fieldnames)
            if minimum_field_names.issubset(file_field_names):
                field_names_to_extract.extend(list(minimum_field_names))
            else:
                error_msg = ("The TSV file given should contain a header that"
                             " has at minimum the following fields "
                             f"{minimum_field_names}. The field names found "
                             f"were {file_field_names}")
                raise ValueError(error_msg)
            
            for extra_field_name in extra_field_names:
                if extra_field_name in file_field_names:
                    field_names_to_extract.append(extra_field_name)
            
            for row in csv_reader:
                row_data: typing.MutableMapping[str, Union[str, List[str]]] = {}
                for field_name in field_names_to_extract:
                    if field_name == 'semantic_tags':
                        row_data[field_name] = row[field_name].split()
                    else:
                        row_data[field_name] = row[field_name]
                collection_from_tsv.add_lexicon_entry(LexiconEntry(**row_data),
                                                      include_pos=include_pos)
        
        return collection_from_tsv.to_dictionary()

    def __setitem__(self, key: str, value: List[str]) -> None:
        self.data[key] = value

    def __getitem__(self, key: str) -> List[str]:
        return self.data[key]

    def __delitem__(self, key: str) -> None:
        del self.data[key]

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> Generator[str, None, None]:
        for key in self.data:
            yield key

    def __str__(self) -> str:
        '''
        Human readable string.
        '''
        object_str = 'LexiconCollection('
        for index, item in enumerate(self.items()):
            object_str += f"('{item[0]}': {item[1]}), "
            if index == 1:
                object_str += '... '
                break
        object_str += f') ({len(self)} entires in the collection)'
        return object_str

    def __repr__(self) -> str:
        '''
        Machine readable string. When printed and run `eval()` over the string
        you should be able to recreate the object.
        '''
        return f'{self.__class__.__name__}(data={self.data})'

    def __eq__(self, other: object) -> bool:
        '''
        Given another object to compare too it will return `True` if the other
        object is the same class and contains the same `data` instance attribute.

        # Parameters

        other : `object`
            The object to compare too.
        
        # Returns

        `True`
        '''
        if not isinstance(other, LexiconCollection):
            return False
        
        if len(self) != len(other):
            return False
        
        if self.data != other.data:
            return False
        
        return True


class MWELexiconCollection(MutableMapping):
    r'''
    A collection that stores Multi Word Expression (MWE) templates and their
    associated meta data.
    
    This collection allows users to:

    1. Easily load MWE templates from a single TSV file.
    2. Find strings that match MWE templates taking into account
    any special syntax rules that should be applied, e.g. wildcards allow zero
    or more characters to appear after the word token and/or Part Of Speech (POS) tag.
    For more information on the MWE special syntax rules see the following
    [notes](/usage/notes/mwe_syntax).
    3. POS mapping, it can find strings that match MWE templates while taking
    into account mapping from one POS tagset to another in both a one to one and
    one to many mapping.

    **Note** that even though this a sub-class of a MutableMapping it has a
    time complexity of O(n) for deletion unlike the standard Python MutableMapping,
    see the [following dict time complexities](https://wiki.python.org/moin/TimeComplexity),
    this is due to keeping track of the `longest_non_special_mwe_template` and
    `longest_wildcard_mwe_template`.

    As we do not currently support curly braces MWE template syntax, therefore
    any MWE templates that contain a `{` or `}` will be ignored and will not be
    added to this collection, in addition a `UserWarning` will be raised stating
    this.

    # Parameters

    data: `Dict[str, List[str]]`, optional (default = `None`)
        Dictionary where the keys are MWE templates, of any :class:`LexiconType`,
        and the values are a list of associated semantic tags.
    pos_mapper : `Dict[str, List[str]]`, optional (default = `None`)
        If not `None`, maps from the lexicon's POS tagset to the desired
        POS tagset, whereby the mapping is a `List` of tags, at the moment there
        is no preference order in this list of POS tags. The POS mapping is
        useful in situtation whereby the leixcon's POS tagset is different to
        the token's. **Note** that the longer the `List[str]` for each POS
        mapping the longer it will take to match MWE templates. A one to one
        mapping will have no speed impact on the tagger. A selection of POS
        mappers can be found in :mod:`pymusas.pos_mapper`.

    # Instance Attributes
    
    **Note** if the `data` parameter given was `None` then the value of all
    dictionary attributes will be an empty dictionary and all integer values will
    be `0`. If `pos_mapper` parameter was `None` then the `pos_mapper` attribute
    will be an empty dictionary.
    
    meta_data: `Dict[str, LexiconMetaData]`
        Dictionary where the keys are MWE templates, of any type, and the values
        are their associated meta data stored in a :class:`LexiconMetaData` object.
    longest_non_special_mwe_template : `int`
        The longest MWE template with no special symbols measured by n-gram size.
        For example the MWE template `ski_noun boot_noun` will be of length 2.
    longest_wildcard_mwe_template : `int`
        The longest MWE template with at least one wildcard (`*`) measured by n-gram size.
        For example the MWE template `*_noun boot*_noun` will be of length 2.
    longest_mwe_template : `int`
        The longest MWE template regardless of type measured by n-gram size.
    most_wildcards_in_mwe_template : `int`
        The number of wildcards in the MWE template that contains the
        most wildcards, e.g. the MWE template `ski_* *_noun` would contain 2
        wildcards. This can be 0 if you have no wildcard MWE templates.
    mwe_regular_expression_lookup: `Dict[int, Dict[str, Dict[str, re.Pattern]]]`
        A dictionary that can lookup all special syntax MWE templates there
        regular expression pattern. These templates are found first by
        their n-gram length and then their first character symbol. The regular
        expression pattern is used for quick matching within the :func:`mwe_match`.
        From the special syntax only wildcard (`*`) symbols are supported at the
        moment.
    pos_mapper : `Dict[str, List[str]]`
        The given `pos_mapper`.
    one_to_many_pos_tags : `Set[str]`
        A set of POS tags that have a one to many mapping, this is created based
        on the `pos_mapper`. This is empty if `pos_mapper` is `None`
    pos_mapping_lookup : `Dict[str, str]`
        Only used if `pos_mapper` is not `None`. For all one-to-one POS mappings
        will store the mapped POS MWE template as keys and the original non-mapped
        (original) MWE templates as values, which can be used to lookup the meta
        data from `meta_data`.
    pos_mapping_regular_expression_lookup : `Dict[LexiconType, Dict[int, Dict[str, Dict[str, re.Pattern]]]]`
        Only used if `pos_mapper` is not `None` and will result in
        `mwe_regular_expression_lookup` being empty as it replaces it
        functionality and extends it and by handlining the one-to-many POS
        mapping cases. When we have a one-to-many POS mapping case this requires
        a regular expression mapping even for non special syntax MWE templates.
        Compared to the `mwe_regular_expression_lookup` the first set of keys
        represent the lexicon entry match type.

    # Examples
    ``` python
    >>> import re
    >>> from pymusas.lexicon_collection import MWELexiconCollection, LexiconType
    >>> mwe_collection = MWELexiconCollection()
    >>> mwe_collection['*_noun boot*_noun'] = ['Z0', 'Z3']
    >>> meta_data = mwe_collection['*_noun boot*_noun']
    >>> assert 2 == meta_data.n_gram_length
    >>> assert LexiconType.MWE_WILDCARD == meta_data.lexicon_type
    >>> assert 2 == meta_data.wildcard_count
    >>> most_likely_tag = meta_data.semantic_tags[0]
    >>> assert most_likely_tag == 'Z0'
    >>> least_likely_tag = meta_data.semantic_tags[-1]
    >>> assert least_likely_tag == 'Z3'
    >>> # change defaultdict to dict so the dictionary is easier to read and understand
    >>> assert ({k: dict(v) for k, v in mwe_collection.mwe_regular_expression_lookup.items()}
    ...         == {2: {'*': {'*_noun boot*_noun': re.compile('[^\\s_]*_noun\\ boot[^\\s_]*_noun')}}})
    
    ```

    '''
    
    def __init__(self, data: Optional[Dict[str, List[str]]] = None,
                 pos_mapper: Optional[Dict[str, List[str]]] = None) -> None:

        self.meta_data: Dict[str, LexiconMetaData] = {}
        self.longest_non_special_mwe_template = 0
        self.longest_wildcard_mwe_template = 0
        self.longest_mwe_template = 0
        self.most_wildcards_in_mwe_template = 0
        self.mwe_regular_expression_lookup: DefaultDict[int, DefaultDict[str, Dict[str, re.Pattern]]]\
            = collections.defaultdict(lambda: collections.defaultdict(dict))

        self.pos_mapper: Dict[str, List[str]] = {}
        self.one_to_many_pos_tags: Set[str] = set()
        self.pos_mapping_lookup: Dict[str, str] = {}
        self.pos_mapping_regular_expression_lookup: DefaultDict[LexiconType, DefaultDict[int, DefaultDict[str, Dict[str, re.Pattern]]]]\
            = collections.defaultdict(lambda: collections.defaultdict(lambda: collections.defaultdict(dict)))
        
        if pos_mapper is not None:
            self.pos_mapper = pos_mapper
            for from_pos, to_pos in pos_mapper.items():
                if len(to_pos) > 1:
                    self.one_to_many_pos_tags.add(from_pos)
        
        if data is not None:
            for key, value in data.items():
                self[key] = value

    def mwe_match(self, mwe_template: str, mwe_type: LexiconType
                  ) -> List[str]:
        '''
        Returns a `List` of MWE templates, with the given `mwe_type`, that match
        the given `mwe_template`. If there are no matches the returned `List`
        will be empty.
        
        This method applies all of the special syntax rules that should be applied
        e.g. wildcards allow zero or more characters to appear after the word
        token and/or Part Of Speech (POS) tag. For more information on the MWE
        special syntax rules see the following [notes](/usage/notes/mwe_syntax).

        # Parameters
        
        mwe_template : `str`
            The MWE template that you want to match against, e.g.
            `river_noun bank_noun` or `ski_noun boots_noun`
        mwe_type : `LexiconType`
            The type of MWE templates that you want to return.

        # Returns

        `Optional[List[str]]`

        # Examples
        ``` python
        >>> from pymusas.lexicon_collection import MWELexiconCollection, LexiconType
        >>> collection = MWELexiconCollection({'walking_noun boot_noun': ['Z2'], 'ski_noun boot_noun': ['Z2'], '*_noun boot_noun': ['Z2'], '*_noun *_noun': ['Z2']})
        >>> assert [] == collection.mwe_match('river_noun bank_noun', LexiconType.MWE_NON_SPECIAL)
        >>> assert ['walking_noun boot_noun'] == collection.mwe_match('walking_noun boot_noun', LexiconType.MWE_NON_SPECIAL)
        >>> assert ['*_noun boot_noun', '*_noun *_noun'] == collection.mwe_match('walking_noun boot_noun', LexiconType.MWE_WILDCARD)
        
        ```
        '''
        mwe_templates_matches: List[str] = []
        if self.pos_mapper:
            if mwe_type == LexiconType.MWE_NON_SPECIAL:
                mwe_mapped_template = self.pos_mapping_lookup.get(mwe_template, None)
                if mwe_mapped_template is not None:
                    potential_match = self.meta_data.get(mwe_mapped_template, None)
                    if potential_match is not None:
                        potential_match_type = potential_match.lexicon_type
                        if LexiconType.MWE_NON_SPECIAL == potential_match_type:
                            mwe_templates_matches.append(mwe_mapped_template)
            
            requires_regular_expression_matching = False
            if mwe_type == LexiconType.MWE_WILDCARD:
                requires_regular_expression_matching = True
            if not mwe_templates_matches and self.one_to_many_pos_tags:
                requires_regular_expression_matching = True
            if requires_regular_expression_matching:
                n_gram_length = len(mwe_template.split())
                mwe_template_length = len(mwe_template)
                if mwe_template_length > 0:
                    for character_lookup in ['*', mwe_template[0]]:
                        regular_expression_lookup = self.pos_mapping_regular_expression_lookup[mwe_type][n_gram_length]
                        if character_lookup not in regular_expression_lookup:
                            continue

                        for (potential_mwe_match,
                             mwe_pattern) in regular_expression_lookup[character_lookup].items():
                            match = mwe_pattern.match(mwe_template)
                            if match is not None:
                                if (match.start() == 0
                                   and match.end() == mwe_template_length):
                                    mwe_templates_matches.append(potential_mwe_match)
        else:
            if mwe_type == LexiconType.MWE_NON_SPECIAL:
                potential_match = self.meta_data.get(mwe_template, None)
                if potential_match is not None:
                    potential_match_type = potential_match.lexicon_type
                    if LexiconType.MWE_NON_SPECIAL == potential_match_type:
                        mwe_templates_matches.append(mwe_template)
            elif mwe_type == LexiconType.MWE_WILDCARD:
                n_gram_length = len(mwe_template.split())
                mwe_template_length = len(mwe_template)
                if mwe_template_length > 0:
                    # By default all MWE matches can start with a * as it covers all characters.
                    for character_lookup in ['*', mwe_template[0]]:
                        regular_expression_lookup = self.mwe_regular_expression_lookup[n_gram_length]
                        if character_lookup not in regular_expression_lookup:
                            continue

                        for (potential_mwe_match,
                             mwe_pattern) in regular_expression_lookup[character_lookup].items():
                            match = mwe_pattern.match(mwe_template)
                            if match is not None:
                                if (match.start() == 0
                                   and match.end() == mwe_template_length):
                                    mwe_templates_matches.append(potential_mwe_match)

        return mwe_templates_matches

    def to_dictionary(self) -> Dict[str, List[str]]:
        '''
        Returns a dictionary of all MWE templates, the keys, stored in the
        collection and their associated semantic tags, the values.

        This can then be used to re-create a :class:`MWELexiconCollection`.

        # Returns

        `Dict[str, List[str]]`

        # Examples
        ``` python
        >>> from pymusas.lexicon_collection import (MWELexiconCollection,
        ... LexiconType, LexiconMetaData)
        >>> mwe_collection = MWELexiconCollection()
        >>> mwe_collection['*_noun boot*_noun'] = ['Z0', 'Z3']
        >>> assert (mwe_collection['*_noun boot*_noun']
        ... == LexiconMetaData(['Z0', 'Z3'], 2, LexiconType.MWE_WILDCARD, 2))
        >>> assert (mwe_collection.to_dictionary()
        ... == {'*_noun boot*_noun': ['Z0', 'Z3']})
        
        ```
        '''
        
        return {key: value.semantic_tags for key, value in self.items()}

    def to_bytes(self) -> bytes:
        '''
        Serialises the :class:`MWELexiconCollection` to a bytestring.

        # Returns

        `bytes`
        '''
        serialise = {}
        data: Dict[str, List[str]] = {key: value.semantic_tags
                                      for key, value in self.meta_data.items()}
        serialise['data'] = srsly.msgpack_dumps(data)
        serialise['pos_mapper'] = srsly.msgpack_dumps(self.pos_mapper)
        return cast(bytes, srsly.msgpack_dumps(serialise))

    @staticmethod
    def from_bytes(bytes_data: bytes) -> "MWELexiconCollection":
        '''
        Loads :class:`MWELexiconCollection` from the given bytestring and
        returns it.

        # Parameters

        bytes_data : `bytes`
            The bytestring to load.
        
        # Returns

        :class:`MWELexiconCollection`
        '''
        serialise_data = srsly.msgpack_loads(bytes_data)
        data = srsly.msgpack_loads(serialise_data['data'])
        pos_mapper = srsly.msgpack_loads(serialise_data['pos_mapper'])
        return MWELexiconCollection(data, pos_mapper)

    @staticmethod
    def from_tsv(tsv_file_path: Union[PathLike, str]
                 ) -> Dict[str, List[str]]:
        '''
        Given a `tsv_file_path` it will return a dictionary object
        that can be used to create a :class:`MWELexiconCollection`.

        Each line in the TSV file will be read in and added to a temporary
        :class:`MWELexiconCollection`, once all lines
        in the TSV have been parsed, the return value is the `data` attribute of
        the temporary :class:`MWELexiconCollection`.

        If the file path is a URL, the file will be downloaded and cached using
        :func:`pymusas.file_utils.download_url_file`.

        Code reference, the identification of a URL and the idea to do this has
        come from the [AllenNLP library](https://github.com/allenai/allennlp/blob/main/allennlp/common/file_utils.py#L205)

        # Parameters

        tsv_file_path: `Union[PathLike, str]`
            A file path or URL to a TSV file that contains at least these two
            fields:
            
            1. `mwe_template`,
            2. `semantic_tags`
            
            All other fields will be ignored.

        # Returns
        
        `Dict[str, List[str]]`
        
        # Raises
        
        `ValueError`
            If the minimum field headings, `mwe_template` and `semantic_tags`,
            do not exist in the given TSV file.

        # Examples

        ``` python
        >>> from pymusas.lexicon_collection import MWELexiconCollection
        >>> portuguese_lexicon_url = 'https://raw.githubusercontent.com/UCREL/Multilingual-USAS/master/Portuguese/mwe-pt.tsv'
        >>> mwe_lexicon_dict = MWELexiconCollection.from_tsv(portuguese_lexicon_url)
        >>> mwe_lexicon_collection = MWELexiconCollection(mwe_lexicon_dict)
        >>> assert mwe_lexicon_dict['abaixo_adv de_prep'][0] == 'M6'
        >>> assert mwe_lexicon_dict['arco_noun e_conj flecha_noun'][0] == 'K5.1'

        ```

        '''
        minimum_field_names = {'mwe_template', 'semantic_tags'}

        collection_from_tsv = MWELexiconCollection()

        if not isinstance(tsv_file_path, str):
            tsv_file_path = str(tsv_file_path)

        parsed = urlparse(tsv_file_path)
        if parsed.scheme in ("http", "https", "s3", "hf", "gs"):
            tsv_file_path = file_utils.download_url_file(tsv_file_path)

        with open(tsv_file_path, 'r', newline='', encoding='utf-8') as fp:
            csv_reader = csv.DictReader(fp, delimiter='\t')
            file_field_names: Set[str] = set()
            if csv_reader.fieldnames:
                file_field_names = set(csv_reader.fieldnames)
            if not minimum_field_names.issubset(file_field_names):
                error_msg = (f"The TSV file, {tsv_file_path}, given should "
                             "contain a header that"
                             " has at minimum the following fields "
                             f"{minimum_field_names}. The field names found "
                             f"were {file_field_names}")
                raise ValueError(error_msg)
            
            for row in csv_reader:
                mwe_template = ''
                semantic_tags: List[str] = []
                for field_name in minimum_field_names:
                    if field_name == 'semantic_tags':
                        semantic_tags = row[field_name].split()
                    elif field_name == 'mwe_template':
                        mwe_template = row[field_name]
                collection_from_tsv[mwe_template] = semantic_tags
        
        return collection_from_tsv.to_dictionary()

    @staticmethod
    def escape_mwe(mwe_template: str) -> str:
        r'''
        Returns the MWE template escaped so that it can be used in a regular
        expression.
        
        The difference between this and the normal `re.escape`
        method, is that we apply the `re.escape` method to the tokens in the
        MWE template and then replace `\*` with `[^\s_]*` so that the wildcards
        keep there original meaning with respect to the MWE special syntax rules.
        Furthermore, the POS tags in the MWE template replace the `*` with
        `[^\s_]*`.

        # Parameters

        mwe_template : `str`
            The MWE template that you want to escape, e.g.
            `river_noun bank_noun` or `*_noun boot*_noun`

        # Returns

        `str`

        # Examples
        ``` python
        >>> from pymusas.lexicon_collection import MWELexiconCollection
        >>> mwe_escaped = MWELexiconCollection.escape_mwe('ano*_prep carta_noun')
        >>> assert r'ano[^\s_]*_prep\ carta_noun' == mwe_escaped
        >>> mwe_escaped = MWELexiconCollection.escape_mwe('ano_prep carta_*')
        >>> assert r'ano_prep\ carta_[^\s_]*' == mwe_escaped

        ```
        '''
        
        escaped_mwe_template_list: List[str] = []
        for token, pos in utils.token_pos_tags_in_lexicon_entry(mwe_template):
            escaped_token = re.escape(token).replace(r'\*', r'[^\s_]*')
            escaped_pos = pos.replace(r'*', r'[^\s_]*')
            escaped_mwe_template_list.append(f'{escaped_token}_{escaped_pos}')
        escaped_mwe_template = r'\ '.join(escaped_mwe_template_list)
        return escaped_mwe_template
    
    def __setitem__(self, key: str, value: List[str]) -> None:
        '''
        # Raises

        `ValueError`
            If using a `pos_mapper` all POS tags within a MWE template cannot
            contain any wildcards or the POS tags can only be a wildcard, if
            this is not the case a `ValueError` will be raised.
        '''
        if '{' in key or '}' in key:
            warnings.warn('We do not currently support Curly Braces expressions'
                          ' within Multi Word Expression (MWE) lexicons and '
                          'therefore any MWE template that contains a `{` '
                          'or `}` will be ignored.')
            return None

        semantic_tags = value
        key_n_gram_length = len(key.split())
        mwe_type: LexiconType = LexiconType.MWE_NON_SPECIAL
        wildcard_count = 0
        
        if not self.pos_mapper:
            if '*' in key:
                mwe_type = LexiconType.MWE_WILDCARD
                wildcard_count += key.count('*')

                if wildcard_count > self.most_wildcards_in_mwe_template:
                    self.most_wildcards_in_mwe_template = wildcard_count
                
                if key_n_gram_length > self.longest_wildcard_mwe_template:
                    self.longest_wildcard_mwe_template = key_n_gram_length
                
                key_as_pattern = re.compile(self.escape_mwe(key))
                self.mwe_regular_expression_lookup[key_n_gram_length][key[0]][key] = key_as_pattern
            else:
                
                if key_n_gram_length > self.longest_non_special_mwe_template:
                    self.longest_non_special_mwe_template = key_n_gram_length
        else:
            contains_one_to_many_pos_tag = False
            pos_mapped_key_list: List[str] = []
            for token, pos in utils.token_pos_tags_in_lexicon_entry(key):
                if '*' in pos:
                    pos_error = ('When using a POS mapper a POS tag within '
                                 'a lexicon entry cannot contain a wildcard'
                                 ' unless the POS tag is only a wildcard '
                                 'and no other characters. Leixcon entry '
                                 'and POS tag in that entry that caused '
                                 f'this error: {key} {pos}')
                    if set(pos.strip()) != set('*'):
                        raise ValueError(pos_error)

                mapped_pos_list = self.pos_mapper.get(pos, [pos])
                if len(mapped_pos_list) > 1:
                    contains_one_to_many_pos_tag = True
                    mapped_pos = '|'.join(mapped_pos_list)
                    mapped_pos = '(?:' + mapped_pos + ')'
                    pos_mapped_key_list.append(f'{token}_{mapped_pos}')
                else:
                    mapped_pos = mapped_pos_list[0]
                    pos_mapped_key_list.append(f'{token}_{mapped_pos}')
            pos_mapped_key = ' '.join(pos_mapped_key_list)
            
            if '*' in pos_mapped_key:
                mwe_type = LexiconType.MWE_WILDCARD
                wildcard_count += key.count('*')

                if wildcard_count > self.most_wildcards_in_mwe_template:
                    self.most_wildcards_in_mwe_template = wildcard_count
                
                if key_n_gram_length > self.longest_wildcard_mwe_template:
                    self.longest_wildcard_mwe_template = key_n_gram_length
                
                key_as_pattern = re.compile(self.escape_mwe(pos_mapped_key))
                self.pos_mapping_regular_expression_lookup[mwe_type][key_n_gram_length][key[0]][key] = key_as_pattern
            elif contains_one_to_many_pos_tag:
                key_as_pattern = re.compile(self.escape_mwe(pos_mapped_key))
                self.pos_mapping_regular_expression_lookup[mwe_type][key_n_gram_length][key[0]][key] = key_as_pattern
                
                if key_n_gram_length > self.longest_non_special_mwe_template:
                    self.longest_non_special_mwe_template = key_n_gram_length
            else:
                self.pos_mapping_lookup[pos_mapped_key] = key
                if key_n_gram_length > self.longest_non_special_mwe_template:
                    self.longest_non_special_mwe_template = key_n_gram_length

        self.longest_mwe_template = max(self.longest_non_special_mwe_template,
                                        self.longest_wildcard_mwe_template)
        self.meta_data[key] = LexiconMetaData(semantic_tags, key_n_gram_length,
                                              mwe_type, wildcard_count)

    def __getitem__(self, key: str) -> LexiconMetaData:
        return self.meta_data[key]

    def __delitem__(self, key: str) -> None:
        
        def _get_lexicon_statistics() -> Tuple[int, int, int]:
            '''
            Returns the `longest_non_special_mwe_template`,
            `longest_wildcard_mwe_template`, and `most_wildcards_in_mwe_template`
            in the `meta_data` as a `Tuple`. This is required as after deleting
            an MWE we do not know if it has affected any of these statistics.

            # Returns

            `Tuple[int, int, int]`
            '''
            longest_non_special_mwe_template = 0
            longest_wildcard_mwe_template = 0
            wildcard_count = 0
            for value in self.values():
                mwe_type = value.lexicon_type
                key_n_gram_length = value.n_gram_length
                if mwe_type == LexiconType.MWE_NON_SPECIAL:
                    if key_n_gram_length > longest_non_special_mwe_template:
                        longest_non_special_mwe_template = key_n_gram_length
                elif mwe_type == LexiconType.MWE_WILDCARD:
                    if key_n_gram_length > longest_wildcard_mwe_template:
                        longest_wildcard_mwe_template = key_n_gram_length
                if value.wildcard_count > wildcard_count:
                    wildcard_count = value.wildcard_count
            return (longest_non_special_mwe_template,
                    longest_wildcard_mwe_template,
                    wildcard_count)
        
        lexicon_meta_data = self[key]
        del self.meta_data[key]
        
        lexicon_type = lexicon_meta_data.lexicon_type
        n_gram_length = lexicon_meta_data.n_gram_length
        if self.pos_mapper:
            if lexicon_type == LexiconType.MWE_WILDCARD:
                del self.pos_mapping_regular_expression_lookup[lexicon_type][n_gram_length][key[0]][key]
            if lexicon_type == LexiconType.MWE_NON_SPECIAL:
                unique_pos_tags_in_key = utils.unique_pos_tags_in_lexicon_entry(key)
                if self.one_to_many_pos_tags.intersection(unique_pos_tags_in_key):
                    del self.pos_mapping_regular_expression_lookup[lexicon_type][n_gram_length][key[0]][key]
                else:
                    _key_to_delete = ''
                    for _key, value in self.pos_mapping_lookup.items():
                        if value == key:
                            _key_to_delete = _key
                    if _key_to_delete:
                        del self.pos_mapping_lookup[_key_to_delete]
        else:
            if lexicon_type == LexiconType.MWE_WILDCARD:
                del self.mwe_regular_expression_lookup[n_gram_length][key[0]][key]
        
        (self.longest_non_special_mwe_template,
         self.longest_wildcard_mwe_template,
         self.most_wildcards_in_mwe_template) = _get_lexicon_statistics()
        self.longest_mwe_template = max(self.longest_non_special_mwe_template,
                                        self.longest_wildcard_mwe_template)

    def __len__(self) -> int:
        return len(self.meta_data)

    def __iter__(self) -> Generator[str, None, None]:
        for key in self.meta_data:
            yield key

    def __str__(self) -> str:
        '''
        Human readable string.
        '''

        object_str = f'{self.__class__.__name__}('
        for index, item in enumerate(self.items()):
            mwe_template = item[0]
            meta_data = item[1]
            object_str += f"('{mwe_template}': {meta_data}), "
            if index == 1:
                object_str += '... '
                break
        object_str += f') ({len(self)} entires in the collection)'
        if self.pos_mapper:
            object_str += ' (Using a POS Mapper)'
        return object_str

    def __repr__(self) -> str:
        '''
        Machine readable string. When printed and run `eval()` over the string
        you should be able to recreate the object.
        '''

        return (f'{self.__class__.__name__}(data={self.to_dictionary()}, '
                f'pos_mapper={self.pos_mapper})')

    def __eq__(self, other: object) -> bool:
        '''
        Given another object to compare too it will return `True` if the other
        object is the same class and contains the same `meta_data` and
        `pos_mapper` instance attributes.

        # Parameters

        other : `object`
            The object to compare too.
        
        # Returns

        `True`
        '''
        if not isinstance(other, MWELexiconCollection):
            return False
        
        if len(self) != len(other):
            return False

        if self.pos_mapper != other.pos_mapper:
            return False

        if self.meta_data != other.meta_data:
            return False
        
        return True



from dataclasses import dataclass
from typing import Tuple

from pymusas.lexicon_collection import LexiconType
from pymusas.rankers.lexical_match import LexicalMatch


@dataclass(init=True, repr=True, eq=True, order=False,
           unsafe_hash=False, frozen=True)
class RankingMetaData:
    '''
    A RankingMetaData object contains all of the meta data about a lexicon
    entry match during the tagging process. This meta data can then be used
    to determine the ranking of the match comapred to other matches within the
    same text/sentence that is being tagged.

    # Instance Attributes

    lexicon_type : `LexiconType`
        Type associated to the lexicon entry.
    lexicon_n_gram_length : `int`
        The n-gram size of the lexicon entry, e.g. `*_noun boot*_noun` will be
        of length 2 and all single word lexicon entries will be of length 1.
    lexicon_wildcard_count : `int`
        Number of wildcards in the lexicon entry, e.g. `*_noun boot*_noun` will
        be 2 and `ski_noun boot_noun` will be 0.
    exclude_pos_information : `bool`
        Whether the POS information was excluded in the match. This is only `True`
        when the match ignores the POS information for single word lexicon entries.
        This is always `False` when used in a Multi Word Expression (MWE) lexicon
        entry match.
    lexical_match : `LexicalMatch`
        What :class:`LexicalMatch` the lexicon entry matched on.
    token_match_start_index : `int`
        Index of the first token in the lexicon entry match.
    token_match_end_index : `int`
        Index of the last token in the lexicon entry match.
    lexicon_entry_match : `str`
        The lexicon entry match, which can be either a single word or MWE entry
        match. In the case for single word this could be `Car|noun` and in the
        case for a MWE it would be it's template, e.g. `snow_noun boots_noun`.
    semantic_tags : `Tuple[str, ...]`
        The semantic tags associated with the lexicon entry. The semantic tags
        are in rank order, the most likely tag is the first tag in the tuple.
        The Tuple can be of variable length hence the `...` in the
        type annotation.
    '''
    lexicon_type: LexiconType
    lexicon_n_gram_length: int
    lexicon_wildcard_count: int
    exclude_pos_information: bool
    lexical_match: LexicalMatch
    token_match_start_index: int
    token_match_end_index: int
    lexicon_entry_match: str
    semantic_tags: Tuple[str, ...]



from enum import IntEnum


class LexicalMatch(IntEnum):
    '''
    Descriptions of the lexical matches and their ordering in tagging priority
    during ranking. Lower the value and rank the higher the tagging priority.

    The `value` attribute of each instance attribute is of type `int`. For the
    best explanation see the example below.

    # Instance Attributes

    TOKEN : `int`
        The lexicon entry matched on the token text.
    LEMMA : `int`
        The lexicon entry matched on the lemma of the token.
    TOKEN_LOWER : `int`
        The lexicon entry matched on the lower cased token text.
    LEMMA_LOWER : `int`
        The lexicon entry matched on the lower cased lemma of the token.

    # Examples
    ``` python
    >>> from pymusas.rankers.lexical_match import LexicalMatch
    >>> assert 1 == LexicalMatch.TOKEN
    >>> assert 'TOKEN' == LexicalMatch.TOKEN.name
    >>> assert 1 == LexicalMatch.TOKEN.value
    ...
    >>> assert 2 == LexicalMatch.LEMMA
    >>> assert 3 == LexicalMatch.TOKEN_LOWER
    >>> assert 4 == LexicalMatch.LEMMA_LOWER
    ...
    >>> assert 2 < LexicalMatch.LEMMA_LOWER

    ```
    '''
    TOKEN = 1
    LEMMA = 2
    TOKEN_LOWER = 3
    LEMMA_LOWER = 4

    def __repr__(self) -> str:
        '''
        Machine readable string. When printed and run `eval()` over the string
        you should be able to recreate the object.
        '''
        return self.__str__()



"""
Base classes for custom classes to inherit from.
"""

from abc import ABC, abstractmethod
import importlib
from typing import Iterable, List, cast

import srsly


class Serialise(ABC):
    '''
    An **abstract class** that defines the basic methods, `to_bytes`, and
    `from_bytes` that is required for all :class:`Serialise`s.
    '''
    @abstractmethod
    def to_bytes(self) -> bytes:
        '''
        Serialises the class to a bytestring.

        # Returns

        `bytes`
        '''
        ...  # pragma: no cover

    @staticmethod
    @abstractmethod
    def from_bytes(bytes_data: bytes) -> "Serialise":
        '''
        Loads the class from the given bytestring and returns it.

        # Parameters

        bytes_data : `bytes`
            The bytestring to load.
        
        # Returns

        :class:`Serialise`
        '''
        ...  # pragma: no cover

    @staticmethod
    def serialise_object_to_bytes(serialise_object: "Serialise"
                                  ) -> bytes:
        '''
        Given a serialise object it will serialise it to a bytestring.

        This function in comparison to calling `to_bytes` on the serialise
        object saves meta data about what class it is so that when loading the
        bytes data later on you will know which class saved the data, this
        would not happen if you called `to_bytes` on the custom object.

        # Parameters

        serialise_object : `Serialise`
            The serialise object, of type :class:`Serialise`, to serialise.

        # Returns

        `bytes`
        '''
        class_name = serialise_object.__class__.__name__
        module_name = serialise_object.__module__
        serialised_object = serialise_object.to_bytes()
        serialised_object_and_meta_data \
            = (serialised_object, (class_name, module_name))
        return cast(bytes, srsly.msgpack_dumps(serialised_object_and_meta_data))

    @staticmethod
    def serialise_object_from_bytes(bytes_data: bytes) -> "Serialise":
        '''
        Loads the serialise object from the given bytestring and return it.
        This is the inverse of function of :func:`serialise_object_to_bytes`.

        # Parameters

        bytes_data : `bytes`
            The bytestring to load.

        # Returns

        `Serialise`
        '''
        serialised_custom_object, meta_data = srsly.msgpack_loads(bytes_data)
        class_name, module_name = meta_data
        custom_object_class: Serialise \
            = getattr(importlib.import_module(module_name), class_name)
        return custom_object_class.from_bytes(serialised_custom_object)

    @staticmethod
    def serialise_object_list_to_bytes(serialise_objects: Iterable["Serialise"]
                                       ) -> bytes:
        '''
        Serialises an `Iterable` of serialise objects in the same way as
        :func:`serialise_object_to_bytes`.

        # Parameters

        serialise_objects : `Iterable[Serialise]`
            The serialise objects, of type :class:`Serialise`, to serialise.

        # Returns

        `bytes`
        '''
        serialised_objects: List[bytes] = []
        for serialise_object in serialise_objects:
            serialised_object \
                = Serialise.serialise_object_to_bytes(serialise_object)
            serialised_objects.append(serialised_object)
        return cast(bytes, srsly.msgpack_dumps(serialised_objects))

    @staticmethod
    def serialise_object_list_from_bytes(bytes_data: bytes
                                         ) -> Iterable["Serialise"]:
        '''
        Loads the serialise objects from the given bytestring and return them.
        This is the inverse of function of
        :func:`serialise_object_list_to_bytes`.

        # Parameters

        bytes_data : `bytes`
            The bytestring to load.

        # Returns

        `Iterable[Serialise]`
        '''
        serialised_objects: List[bytes] = srsly.msgpack_loads(bytes_data)
        serialise_objects: List[Serialise] = []
        for serialised_object in serialised_objects:
            serialise_object \
                = Serialise.serialise_object_from_bytes(serialised_object)
            serialise_objects.append(serialise_object)
        return serialise_objects



'''
Helper functions for anything [spaCy](https://spacy.io/) related in the code
base.
'''

import copy
import warnings

from spacy.language import Language
from spacy.pipe_analysis import validate_attrs
from spacy.tokens import Token


def set_custom_token_extension(extension_name: str) -> None:
    '''
    Defines a custom attribute of the spaCy Token which becomes avaliable
    via `Token._.{extension_name}`. The difference between this and using the
    spaCy [Token.set_extension method](https://spacy.io/api/token#set_extension)
    is this method will check if the extension exists already and if so will force it
    through and output an UserWarning message that it has had to force this through.

    # Parameters

    extension_name : `str`
        Name of the custom attribute that will become avaliable through
        `Token._.{extension_name}`.
    '''
    if Token.has_extension(extension_name):
        old_extension = Token.get_extension(extension_name)
        Token.set_extension(extension_name, default=None, force=True)
        message = (f'Overwritten the spaCy Token extension `{extension_name}`'
                   ' which currently has the following (default, method, getter, setter):'
                   f'`{old_extension}`. And replacing it with the following:'
                   f'`{Token.get_extension(extension_name)}`'
                   '. This would only become a problem if the the two Tuples'
                   ' of four are different, if they are the same there is'
                   ' no problem.')
        warnings.warn(message)
    else:
        Token.set_extension(extension_name, default=None)


def update_factory_attributes(meta_information_to_update: str,
                              factory_name: str,
                              new_attribute_name: str,
                              old_attribute_name: str) -> None:
    '''
    Updates the
    [spaCy Language meta information](https://spacy.io/api/language#factorymeta)
    for either `assigns` or `requires` for the given component, find through
    it's factory name, by replacing the `old_attribute_name` with the
    `new_attribute_name`.

    # Parameters

    meta_information_to_update : `str`
        Either `assigns` or `requires`, raises a ValueError if it is any other
        value.
    factory_name : `str`
        The name of the component factory, e.g. `pymusas_rule_based_tagger`
    new_attribute_name : `str`
        The name of the new attribute that is required for this component.
        An example, `token.pos`.
    old_attribute_name : `str`
        The name of the old attribute that is to be replaced with
        the `new_attribute_name`. An example, `token.tag`.
    '''
    value_error = ('`meta_information_to_update` has to be either `assigns` '
                   f'or `requires` and not {meta_information_to_update}')
    if meta_information_to_update not in set(['assigns', 'requires']):
        raise ValueError(value_error)

    factory_meta = Language.get_factory_meta(factory_name)
    required_attributes = copy.deepcopy(getattr(factory_meta,
                                                meta_information_to_update))
    updated_attributes = [attribute for attribute in required_attributes
                          if attribute != old_attribute_name]
    updated_attributes.append(f'{new_attribute_name}')
    
    setattr(factory_meta, meta_information_to_update,
            validate_attrs(updated_attributes))
